import logging
from pathlib import Path

import dask.dataframe as dd
import geopandas as gpd
import numpy as np
import pandas as pd
from dask import compute, delayed
from dask.diagnostics import ProgressBar
from pandas.api.types import is_string_dtype
from spatialdata import SpatialData

from .. import settings
from ..constants import SopaFiles, SopaKeys
from ..spatial import assign_transcript_to_cell
from ..utils import add_spatial_element, get_cache_dir, get_feature_key, to_intrinsic
from .patches import Patches2D

log = logging.getLogger(__name__)


class OnDiskTranscriptPatches(Patches2D):
    def __init__(
        self,
        sdata: SpatialData,
        points_key: str,
        patch_width: float,
        patch_overlap: float,
        prior_shapes_key: str | None = None,
        unassigned_value: int | str | None = None,
        min_points_per_patch: int = 4000,
        min_cells_per_patch: int = 1,
        csv_name: str = SopaFiles.TRANSCRIPTS_FILE,
        centroids_csv_name: str = SopaFiles.CENTROIDS_FILE,
        write_cells_centroids: bool = False,
        roi_key: str | None = SopaKeys.ROI,
    ):
        super().__init__(sdata, points_key, patch_width, patch_overlap, roi_key=roi_key)

        self.points_key = points_key
        self.points: dd.DataFrame = sdata.points[points_key]

        assert isinstance(self.points, dd.DataFrame), (
            "Points should be a dask DataFrame. Please report this issue on Sopa's Github repository."
        )

        if "z" not in self.points.columns:
            self.points["z"] = 0  # ensure z column is present - needed for proseg

        self.prior_shapes_key = prior_shapes_key
        self.unassigned_value = unassigned_value

        self.min_points_per_patch = min_points_per_patch
        self.min_cells_per_patch = min_cells_per_patch

        self.csv_name = csv_name
        self.centroids_csv_name = centroids_csv_name
        self.write_cells_centroids = write_cells_centroids

        self.cache_dir = get_cache_dir(sdata) / SopaFiles.TRANSCRIPT_CACHE_DIR

    def write(self):
        self.assign_prior_segmentation()

        self.setup_patches_directory()

        patches_geo_df = self.geo_df.geometry.to_frame()

        if self.write_cells_centroids:
            centroids = self.get_prior_centroids()
            self.write_points(patches_geo_df, centroids, csv_name=self.centroids_csv_name)

        gene_column = get_feature_key(self.points)
        with ProgressBar():
            tasks = [
                delayed(self.query_points_partition)(patches_geo_df, part, gene_column=gene_column)
                for part in self.points.to_delayed()
            ]
            compute(*tasks)

    def assign_prior_segmentation(self) -> None:
        if self.prior_shapes_key is None:
            return

        if self.prior_shapes_key in self.sdata.shapes:
            assert self.unassigned_value is None, (
                "Unassigned value is not needed when using a prior segmentation based on existing shapes"
            )

            assign_transcript_to_cell(
                self.sdata, self.points_key, self.prior_shapes_key, SopaKeys.SOPA_PRIOR, unassigned_value=0
            )
            return

        assert self.prior_shapes_key in self.points.columns, (
            f"Prior-segmentation column {self.prior_shapes_key} not found in sdata['{self.points_key}']"
        )

        self.points[SopaKeys.SOPA_PRIOR] = _unassigned_to_zero(
            self.points[self.prior_shapes_key], self.unassigned_value
        )

    def get_prior_centroids(self) -> gpd.GeoDataFrame:
        assert self.prior_shapes_key is not None, "Prior shapes key is required to write cell centroids"

        assert self.prior_shapes_key in self.sdata.shapes, (
            "When using write_cells_centroids=True, `prior_shapes_key` must be a shape element in `sdata.shapes`."
        )

        centroids = to_intrinsic(self.sdata, self.prior_shapes_key, self.points).geometry.centroid

        return gpd.GeoDataFrame(
            {
                "x": centroids.geometry.x,
                "y": centroids.geometry.y,
                "z": 0,
                SopaKeys.SOPA_PRIOR: range(1, len(centroids) + 1),
            },
            geometry=centroids,
        )

    def add_shapes(self, key_added: str | None = None):
        valid_indices = list(self.valid_indices())

        assert len(valid_indices), "No valid patches found. Check the minimum number of points or cells per patch."

        geo_df = self.geo_df.iloc[valid_indices].copy()
        geo_df[SopaKeys.POINTS_KEY] = self.points_key

        if self.prior_shapes_key:
            geo_df[SopaKeys.PRIOR_SHAPES_KEY] = self.prior_shapes_key

        key_added = key_added or SopaKeys.TRANSCRIPTS_PATCHES

        add_spatial_element(self.sdata, key_added, geo_df)

        log.info(f"Added {len(valid_indices)} patch(es) to sdata['{key_added}']")

    def get_patch_path(self, index: int, csv_name: str | None = None) -> Path:
        return self.cache_dir / str(index) / (csv_name or self.csv_name)

    def setup_patches_directory(self):
        for index in range(len(self)):
            patch_path = self.get_patch_path(index)
            patch_path.parent.mkdir(parents=True, exist_ok=True)
            if patch_path.exists():
                patch_path.unlink()
            pd.DataFrame(columns=self.points.columns).to_csv(patch_path, index=False)

    def valid_indices(self):
        for index in range(len(self)):
            if _check_min_lines(self.get_patch_path(index), self.min_points_per_patch):
                if not self.write_cells_centroids or _check_min_lines(
                    self.get_patch_path(index, csv_name=self.centroids_csv_name), self.min_cells_per_patch
                ):
                    yield index
                    continue

                log.info(f"Patch {index} is out for segmentation (less than {self.min_cells_per_patch} prior cells).")
                continue

            log.info(f"Patch {index} is out for segmentation (less than {self.min_points_per_patch} transcripts).")

    def query_points_partition(
        self, patches_gdf: gpd.GeoDataFrame, df: pd.DataFrame, gene_column: str | None = None
    ) -> pd.DataFrame:
        if SopaKeys.LOW_QUALITY_TRANSCRIPT_KEY in df.columns:
            df = df[~df[SopaKeys.LOW_QUALITY_TRANSCRIPT_KEY]]
        if gene_column is not None and settings.gene_exclude_pattern is not None:
            df = df[~df[gene_column].str.match(settings.gene_exclude_pattern, case=False, na=True)]

        if self.is_full_slide:
            df.to_csv(self.get_patch_path(0), mode="a", header=False, index=False)
        else:
            points_gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df["x"], df["y"]))
            self.write_points(patches_gdf, points_gdf, mode="a")

    def write_points(
        self, patches_gdf: gpd.GeoDataFrame, points_gdf: gpd.GeoDataFrame, mode="w", csv_name: str | None = None
    ):
        patches_gdf.index.name = "index_right"  # to reuse the index name later
        df_merged: gpd.GeoDataFrame = points_gdf.sjoin(patches_gdf)

        for index, patch_df in df_merged.groupby("index_right"):
            patch_path = self.get_patch_path(index, csv_name=csv_name)
            patch_path.parent.mkdir(parents=True, exist_ok=True)
            patch_df = patch_df.drop(columns=["index_right", "geometry"])
            patch_df.to_csv(patch_path, mode=mode, header=mode == "w", index=False)


def _check_min_lines(path: str, n: int) -> bool:
    if not Path(path).exists():  # empty file are not written at all
        return False
    with open(path) as f:
        return any(count >= n for count, _ in enumerate(f))


def _unassigned_to_zero(series: dd.Series, unassigned_value: int | str | None) -> dd.Series:
    if is_string_dtype(series):
        series = series.astype("category")
        series = series.cat.as_known()

        categories = series.cat.categories
        assert unassigned_value in categories, f"Unassigned value {unassigned_value} not in categories"
        categories = categories.delete(categories.get_loc(unassigned_value))
        categories = pd.Index([unassigned_value]).append(categories)

        series = series.cat.reorder_categories(categories)
        return series.cat.codes
    try:
        is_integer_dtype = np.issubdtype(series.dtype, np.integer)
    except:
        is_integer_dtype = False

    if is_integer_dtype:
        if unassigned_value is None or unassigned_value == 0:
            return series
        max_value = np.iinfo(series.dtype).max
        return series.replace(0, max_value - 1).replace(int(unassigned_value), 0)

    raise ValueError(f"Invalid dtype {series.dtype} for prior cell ids. Must be int or string.")



from torch import nn


class Resnet50Features(nn.Module):
    def __init__(self):
        super().__init__()

        try:
            import timm
        except ImportError:
            raise ImportError(
                "Using the 'resnet50' model for inference requires installing the timm dependency, e.g. via `pip install timm`"
            )
        self.model = timm.create_model("resnet50", pretrained=True, num_classes=0)
        self.model.eval()

        from torchvision import transforms

        self.transform = transforms.Compose([
            transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))
        ])

    def __call__(self, x):
        return self.model(self.transform(x))



import logging
from pathlib import Path
from typing import Callable

import dask.array as da
import numpy as np
import pandas as pd
import tifffile as tf
import xarray as xr
from dask_image.imread import imread
from spatialdata import SpatialData
from spatialdata.models import Image2DModel
from xarray import DataArray

from ...constants import SopaAttrs

log = logging.getLogger(__name__)


def _default_image_kwargs(
    image_models_kwargs: dict | None = None, imread_kwargs: dict | None = None
) -> tuple[dict, dict]:
    image_models_kwargs = {} if image_models_kwargs is None else image_models_kwargs
    imread_kwargs = {} if imread_kwargs is None else imread_kwargs

    if "chunks" not in image_models_kwargs:
        image_models_kwargs["chunks"] = (1, 1024, 1024)

    if "scale_factors" not in image_models_kwargs:
        image_models_kwargs["scale_factors"] = [2, 2, 2, 2]

    return image_models_kwargs, imread_kwargs


def _deduplicate_names(names: pd.Series | np.ndarray | list[str]) -> np.ndarray:
    if not isinstance(names, pd.Series):
        names = pd.Series(names)
    names = names.astype(str)

    duplicates = names.duplicated()
    names[duplicates] += " (" + names.groupby(by=names).cumcount().astype(str)[duplicates] + ")"

    return names.values


def _get_ome_channel_names(files):
    return _deduplicate_names([_ome_channels_names(file)[0] for file in files])


def _general_tif_directory_reader(
    path: str,
    files_to_channels: Callable = _get_ome_channel_names,
    suffix: str = ".tif",
    image_models_kwargs: dict | None = None,
    imread_kwargs: dict | None = None,
):
    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)

    files = [file for file in Path(path).iterdir() if file.suffix == suffix]

    names = files_to_channels(files)
    image = da.concatenate(
        [imread(file, **imread_kwargs) for file in files],
        axis=0,
    )
    image = image.rechunk(chunks=image_models_kwargs["chunks"])

    log.info(f"Found channel names {names}")

    image_name = Path(path).absolute().stem
    image = Image2DModel.parse(
        image,
        dims=("c", "y", "x"),
        c_coords=names,
        **image_models_kwargs,
    )

    return SpatialData(images={image_name: image}, attrs={SopaAttrs.CELL_SEGMENTATION: image_name})


def _clip_intensity_values(
    image: xr.DataArray, clip_quantile: bool | None = None, quantile: float = 0.99
) -> xr.DataArray:
    if clip_quantile is False or image.shape[1] * image.shape[2] > 2**28:
        denominator = image.max(dim=("y", "x"))
    else:
        denominator = image.chunk({"y": -1, "x": -1}).quantile(quantile, dim=("y", "x"))

    return ((image / denominator.data.compute()[:, None, None]).clip(0, 1) * 255).astype(np.uint8)


def _image_int_dtype(image: xr.DataArray, clip_quantile: bool | None = None, quantile: float = 0.99) -> xr.DataArray:
    image = image.transpose("c", "y", "x")

    if np.issubdtype(image.dtype, np.integer):
        return image

    if not np.issubdtype(image.dtype, np.floating):
        raise ValueError(f"Invalid image type {image.dtype}")

    return _clip_intensity_values(image, clip_quantile=clip_quantile, quantile=quantile)


def _ome_channels_names(path: Path | str):
    import xml.etree.ElementTree as ET

    tiff = tf.TiffFile(path)
    omexml_string = tiff.pages[0].description

    root = ET.fromstring(omexml_string)
    namespaces = {"ome": "http://www.openmicroscopy.org/Schemas/OME/2016-06"}
    channels = root.findall("ome:Image[1]/ome:Pixels/ome:Channel", namespaces)
    return [c.attrib["Name"] if "Name" in c.attrib else c.attrib["ID"] for c in channels]


def ome_tif(path: str | Path, as_image: bool = False) -> DataArray | SpatialData:
    """Read an `.ome.tif` image. This image should be a 2D image (with possibly multiple channels).
    Typically, this function can be used to open Xenium IF images.

    Args:
        path: Path to the `.ome.tif` image
        as_image: If `True`, will return a `DataArray` object

    Returns:
        A `DataArray` or a `SpatialData` object
    """
    image_models_kwargs, _ = _default_image_kwargs()
    image_name = Path(path).absolute().name.split(".")[0]
    image: da.Array = imread(path)

    if image.ndim == 4:
        assert image.shape[0] == 1, "4D images not supported"
        image = da.moveaxis(image[0], 2, 0)
        log.info(f"Transformed 4D image into a 3D image of shape (c, y, x) = {image.shape}")
    elif image.ndim != 3:
        raise ValueError(f"Number of dimensions not supported: {image.ndim}")

    image = image.rechunk(chunks=image_models_kwargs["chunks"])

    try:
        channel_names = _ome_channels_names(path)
    except:
        channel_names = []
    if len(channel_names) != len(image):
        channel_names = [str(i) for i in range(len(image))]
        log.warning(f"Channel names couldn't be read. Using {channel_names} instead.")

    image = DataArray(image, dims=["c", "y", "x"], name=image_name, coords={"c": channel_names})
    image = _image_int_dtype(image)

    if as_image:
        return image

    image = Image2DModel.parse(
        image,
        c_coords=channel_names,
        **image_models_kwargs,
    )

    return SpatialData(images={image_name: image}, attrs={SopaAttrs.CELL_SEGMENTATION: image_name})



import logging

import numpy as np
import pandas as pd
import zarr
from anndata import AnnData
from scipy.sparse import csr_matrix

from ._constants import FileNames, cell_categories_attrs
from .utils import explorer_file_path, int_cell_id

log = logging.getLogger(__name__)


def write_gene_counts(
    path: str,
    adata: AnnData,
    layer: str | None = None,
    preserve_ids: bool = True,
    is_dir: bool = True,
) -> None:
    """Write a `cell_feature_matrix.zarr.zip` file containing the cell-by-gene transcript counts (i.e., from `adata.X`)

    Args:
        path: Path to the Xenium Explorer directory where the cell-by-gene file will be written
        adata: An `AnnData` object. Note that `adata.X` has to be a sparse matrix (and contain the raw counts), else use the `layer` argument.
        layer: If not `None`, `adata.layers[layer]` should be sparse (and contain the raw counts). By default, look into `adata.layers["counts"]` or use `adata.X`.
        preserve_ids: If `True`, then the cell IDs will be kept in the explorer as they are in `adata.obs_names`. If `False`, then the cell IDs will be replaced by integers starting from 0 in the output file.
        is_dir: If `False`, then `path` is a path to a single file, not to the Xenium Explorer directory.
    """
    path = explorer_file_path(path, FileNames.TABLE, is_dir)

    log.info(f"Writing table with {adata.n_vars} columns")
    counts = adata.layers.get("counts", adata.X) if layer is None else adata.layers[layer]
    counts = csr_matrix(counts.T)

    feature_keys = [*list(adata.var_names), "Total transcripts"]
    feature_ids = feature_keys
    feature_types = ["gene"] * len(adata.var_names) + ["aggregate_gene"]

    ATTRS = {
        "major_version": 3,
        "minor_version": 0,
        "number_cells": adata.n_obs,
        "number_features": adata.n_vars + 1,
        "feature_keys": feature_keys,
        "feature_ids": feature_ids,
        "feature_types": feature_types,
    }

    total_counts = counts.sum(1).A1
    loc = total_counts > 0

    data = np.concatenate([counts.data, total_counts[loc]])
    indices = np.concatenate([counts.indices, np.where(loc)[0]])
    indptr = counts.indptr
    indptr = np.append(indptr, indptr[-1] + loc.sum())

    cell_id = np.ones((adata.n_obs, 2))
    cell_id[:, 0] = int_cell_id(adata.obs_names) if preserve_ids else np.arange(adata.n_obs)

    with zarr.ZipStore(path, mode="w") as store:
        g = zarr.group(store=store)
        cells_group = g.create_group("cell_features")
        cells_group.attrs.put(ATTRS)

        cells_group.array("cell_id", cell_id, dtype="uint32", chunks=cell_id.shape)
        cells_group.array("data", data, dtype="uint32", chunks=data.shape)
        cells_group.array("indices", indices, dtype="uint32", chunks=indices.shape)
        cells_group.array("indptr", indptr, dtype="uint32", chunks=indptr.shape)


def _write_categorical_column(root: zarr.Group, index: int, values: np.ndarray, categories: list[str]) -> None:
    group = root.create_group(index)
    values_indices = [np.where(values == cat)[0] for cat in categories]
    values_cum_len = np.cumsum([len(indices) for indices in values_indices])

    indices = np.concatenate(values_indices)
    indptr = np.concatenate([[0], values_cum_len[:-1]])

    group.array("indices", indices, dtype="uint32", chunks=(len(indices),))
    group.array("indptr", indptr, dtype="uint32", chunks=(len(indptr),))


def write_cell_categories(path: str, adata: AnnData, is_dir: bool = True) -> None:
    """Write a `analysis.zarr.zip` file containing the cell categories/clusters (i.e., from `adata.obs`)

    Args:
        path: Path to the Xenium Explorer directory where the cell-categories file will be written
        adata: An `AnnData` object
        is_dir: If `False`, then `path` is a path to a single file, not to the Xenium Explorer directory.
    """
    path = explorer_file_path(path, FileNames.CELL_CATEGORIES, is_dir)

    adata.strings_to_categoricals()
    cat_columns = [name for name, cat in adata.obs.dtypes.items() if cat == "category"]

    log.info(f"Writing {len(cat_columns)} cell/observations categories: {', '.join(cat_columns)}")

    ATTRS = cell_categories_attrs()
    ATTRS["number_groupings"] = len(cat_columns)

    with zarr.ZipStore(path, mode="w") as store:
        g = zarr.group(store=store)
        cell_groups = g.create_group("cell_groups")

        for i, name in enumerate(cat_columns):
            if adata.obs[name].isna().any():
                NA = "NA"
                log.warning(f"Column {name} has nan values. They will be displayed as '{NA}'")
                adata.obs[name] = adata.obs[name].cat.add_categories(NA).fillna(NA)

            categories = list(adata.obs[name].cat.categories)
            ATTRS["grouping_names"].append(name)
            ATTRS["group_names"].append(categories)

            _write_categorical_column(cell_groups, i, adata.obs[name], categories)

        cell_groups.attrs.put(ATTRS)


def save_column_csv(path: str, adata: AnnData, key: str):
    """Save one column of the AnnData object as a CSV that can be open interactively in the explorer, under the "cell" panel.

    Args:
        path: Path where to write the CSV that will be open in the Xenium Explorer
        adata: An `AnnData` object
        key: Key of `adata.obs` containing the column to convert
    """
    df = pd.DataFrame({"cell_id": adata.obs_names, "group": adata.obs[key].values})
    df.to_csv(path, index=None)



import logging
import re
from pathlib import Path

from spatialdata import SpatialData

from .utils import _deduplicate_names, _general_tif_directory_reader

log = logging.getLogger(__name__)


def macsima(path: Path, **kwargs: int) -> SpatialData:
    """Read MACSIMA data as a `SpatialData` object

    Notes:
        For all dulicated name, their index will be added in brackets after, for instance you may find `DAPI (1)`.

    Args:
        path: Path to the directory containing the MACSIMA `.tif` images
        kwargs: Kwargs for the `_general_tif_directory_reader`

    Returns:
        A `SpatialData` object with a 2D-image of shape `(C, Y, X)`
    """
    files = list(Path(path).glob("*.tif"))

    if any("A-" in file.name for file in files):  # non-ome.tif format
        return _general_tif_directory_reader(path, files_to_channels=_get_channel_names_macsima, **kwargs)

    return _general_tif_directory_reader(path, **kwargs)


def _parse_name_macsima(file):
    match = re.search(r"_A-(.*?)_C-", file.name)
    antibody = match.group(1) if match else re.search(r"_A-(.*?)\.tif", file.name).group(1)
    return antibody


def _get_channel_names_macsima(files):
    return _deduplicate_names([_parse_name_macsima(file) for file in files])



import logging
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import spatialdata
from anndata import AnnData
from scipy.sparse import csr_matrix
from spatialdata import SpatialData
from spatialdata.models import ShapesModel
from spatialdata.transformations import BaseTransformation, Identity

from ...aggregation.aggregation import add_standardized_table
from ...constants import SopaAttrs, SopaKeys
from ...utils import (
    copy_transformations,
    delete_transcripts_patches_dirs,
    get_cache_dir,
    get_feature_key,
    get_transcripts_patches_dirs,
    set_boundaries_attrs,
    to_intrinsic,
)
from .utils import _check_transcript_patches, _get_executable_path

log = logging.getLogger(__name__)


def proseg(
    sdata: SpatialData,
    delete_cache: bool = True,
    command_line_suffix: str = "",
    infer_presets: bool = True,
    prior_shapes_key: str | None = None,
    bins_key: str | None = None,
    key_added: str = SopaKeys.PROSEG_BOUNDARIES,
):
    """Run [`proseg`](https://github.com/dcjones/proseg) segmentation on a SpatialData object, and add the corresponding cell boundaries and `AnnData` table with counts.

    !!! warning "Proseg installation"
        Make sure to install [`proseg`](https://github.com/dcjones/proseg) separately before running this function.

    !!! info "Proseg usage specificities"
        If you are using Visium HD data, provide the `prior_shapes_key` corresponding to the prior cell boundaries (e.g., `"stardist_boundaries"`).

        If you're using non-Visium HD data, note that `proseg` will only run on one patch. I.e., you need
        to run [`sopa.make_transcript_patches`](../patches/#sopa.make_transcript_patches) with `patch_width=None` and a `prior_shapes_key` before running `proseg`.

        Also, note that aggregation is not necessary after running `proseg` (but can be useful if you want e.g. channel aggregation).

    Args:
        sdata: A `SpatialData` object.
        delete_cache: Whether to delete the cache after segmentation.
        command_line_suffix: Optional suffix to add to the proseg command line.
        infer_presets: Whether to infer the proseg presets based on the columns of the transcripts dataframe.
        prior_shapes_key: **Only for Visium HD data.** Key of `sdata` containing the prior cell boundaries. If `"auto"`, use the latest performed segmentation (e.g., stardist or the 10X Genomics segmentation).
        bins_key: **Only for Visium HD data.** Key of `sdata` with the table corresponding to the bin-by-gene table of gene counts (e.g., for Visium HD data). Inferred by default.
        key_added: Name of the shapes element to be added to `sdata.shapes`.
    """
    if prior_shapes_key is not None:
        bins_key = bins_key or sdata.attrs.get(SopaAttrs.BINS_TABLE)

        assert bins_key is not None, (
            "Using `prior_shapes_key` is specific to Visium HD data, and a `bins_key` must be provided."
        )

        _proseg_bins(
            sdata,
            prior_shapes_key=prior_shapes_key,
            bins_key=bins_key,
            command_line_suffix=command_line_suffix,
            infer_presets=infer_presets,
            key_added=key_added,
        )
    else:
        assert sdata.points, (
            "No points found in `sdata`. If you use Visium HD data, consider providing a `prior_shapes_key`."
        )
        assert bins_key is None, (
            "`bins_key` can only be provided when using `prior_shapes_key` (i.e., for Visium HD data)."
        )

        _proseg_points(
            sdata,
            delete_cache=delete_cache,
            command_line_suffix=command_line_suffix,
            infer_presets=infer_presets,
            key_added=key_added,
        )

    log.info("Proseg table and boundaries added (running `sopa.aggregate` is not mandatory).")


def _proseg_points(
    sdata: SpatialData, delete_cache: bool, command_line_suffix: str, infer_presets: bool, key_added: str
):
    _check_transcript_patches(sdata, with_prior=True)

    points_key = sdata[SopaKeys.TRANSCRIPTS_PATCHES][SopaKeys.POINTS_KEY].iloc[0]

    patches_dirs = get_transcripts_patches_dirs(sdata)
    assert len(patches_dirs) == 1, (
        "Proseg is fast enough to work on a single patch. Re-run `sopa.make_transcript_patches` with `patch_width=None` and a `prior_shapes_key`."
    )
    patch_dir = Path(patches_dirs[0])

    proseg_command = _get_proseg_points_command(sdata, points_key, command_line_suffix, infer_presets)

    _run_proseg(proseg_command, patch_dir)
    adata, geo_df = _read_proseg(patch_dir, copy_transformations(sdata[points_key]))

    add_standardized_table(sdata, adata, geo_df, key_added, SopaKeys.TABLE)

    set_boundaries_attrs(sdata, key_added)

    if delete_cache:
        delete_transcripts_patches_dirs(sdata)


def _proseg_bins(
    sdata: SpatialData,
    prior_shapes_key: str,
    bins_key: str,
    command_line_suffix: str,
    infer_presets: bool,
    key_added: str,
):
    import shutil

    if prior_shapes_key == "auto":
        prior_shapes_key = sdata.attrs.get(SopaAttrs.BOUNDARIES)
        assert prior_shapes_key is not None, (
            "`sdata` does not contain any known prior segmentation. Please provide the `prior_shapes_key` explicitly."
        )

    bins_table: AnnData = sdata.tables[bins_key]

    assert isinstance(bins_table.X, csr_matrix), (
        "The bin table's X matrix must be in CSR format. E.g., you can run `adata.X = adata.X.tocsr()`"
    )

    bins_shapes_key = sdata.get_annotated_regions(bins_table)
    bins_shapes_key = bins_shapes_key[0] if isinstance(bins_shapes_key, list) else bins_shapes_key
    bins_shapes = sdata[bins_shapes_key]

    assert "microns" in bins_shapes.attrs["transform"], "`microns` coordinate system is needed for the shapes"

    centroid_microns = spatialdata.transform(bins_shapes, to_coordinate_system="microns").centroid
    bins_table.obsm["spatial_microns"] = np.array([centroid_microns.x, centroid_microns.y]).T

    prior_aligned = to_intrinsic(sdata, prior_shapes_key, bins_shapes_key)

    sjoin = sdata[bins_shapes_key].sjoin(prior_aligned.reset_index(drop=True))
    sjoin = sjoin[~sjoin.index.duplicated()]

    instance_key_column = sdata.get_instance_key_column(bins_table)
    sjoin.index = sjoin.index.map(pd.Series(instance_key_column.index, index=instance_key_column.values))

    bins_table.obs[SopaKeys.SOPA_PRIOR] = 0
    bins_table.obs.loc[sjoin.index, SopaKeys.SOPA_PRIOR] = sjoin["index_right"] + 1

    work_dir = get_cache_dir(sdata)

    bins_table.write_zarr(work_dir / "bins_table.zarr")

    proseg_command = _get_proseg_bins_command(sdata, command_line_suffix, infer_presets)

    _run_proseg(proseg_command, work_dir)

    adata, geo_df = _read_proseg(work_dir, _transformations_from_microns(bins_shapes))

    add_standardized_table(sdata, adata, geo_df, key_added, SopaKeys.TABLE)

    set_boundaries_attrs(sdata, key_added)

    shutil.rmtree(work_dir / "bins_table.zarr")
    shutil.rmtree(work_dir / "proseg-output.zarr")


def _transformations_from_microns(bins_shapes: gpd.GeoDataFrame) -> dict[str, BaseTransformation]:
    transformations = copy_transformations(bins_shapes)
    from_microns = transformations["microns"].inverse()

    transformations = {cs: from_microns.compose_with(t) for cs, t in transformations.items() if cs != "microns"}
    transformations["microns"] = Identity()

    return transformations


def _run_proseg(proseg_command: str, cwd: str | Path):
    import subprocess

    log.info(f"Running proseg with command: `{proseg_command}`")

    result = subprocess.run(
        proseg_command,
        cwd=cwd,
        shell=True,
        capture_output=False,
    )

    if result.returncode != 0:
        raise subprocess.CalledProcessError(
            returncode=result.returncode,
            cmd=proseg_command,
            output=result.stdout,
            stderr=result.stderr,
        )


def _get_proseg_points_command(
    sdata: SpatialData,
    points_key: str,
    command_line_suffix: str,
    infer_presets: bool,
) -> str:
    proseg_executable = _get_executable_path("proseg", ".cargo")

    feature_key = get_feature_key(sdata[points_key], raise_error=True)

    use_zarr = _use_zarr_output(proseg_executable)

    if infer_presets:
        command_line_suffix = _add_presets(command_line_suffix, sdata[points_key].columns)

    return f"{proseg_executable} transcripts.csv -x x -y y -z z --gene-column {feature_key} --cell-id-column {SopaKeys.SOPA_PRIOR} --cell-id-unassigned 0 {'--exclude-spatialdata-transcripts' if use_zarr else ''} {command_line_suffix}"


def _get_proseg_bins_command(sdata: SpatialData, command_line_suffix: str, infer_presets: bool) -> str:
    proseg_executable = _get_executable_path("proseg", ".cargo")

    if infer_presets and "--voxel-size" not in command_line_suffix:
        command_line_suffix += " --voxel-size 2"

    return f"{proseg_executable} --anndata {get_cache_dir(sdata) / 'bins_table.zarr'} --anndata-coordinate-key spatial_microns --overwrite --cell-id-column {SopaKeys.SOPA_PRIOR} {command_line_suffix}"


def _add_presets(command_line_suffix: str, columns: list[str]) -> str:
    if "fov-column" not in command_line_suffix:
        for column in columns:
            if column in ["fov", "fov_name"]:
                command_line_suffix += f" --fov-column {column}"
                break

    if "qv" in columns and "--qv-column" not in command_line_suffix:
        command_line_suffix += " --qv-column qv"

    if "qv" in columns and "--min-qv" not in command_line_suffix:
        command_line_suffix += " --min-qv 20.0"

    if "compartment-column" not in command_line_suffix:
        if "overlaps_nucleus" in columns:
            command_line_suffix += " --compartment-column overlaps_nucleus --compartment-nuclear 1"
        elif "CellComp" in columns:
            command_line_suffix += " --compartment-column CellComp --compartment-nuclear Nuclear"

    return command_line_suffix


def _read_proseg(cwd: Path, transformations: dict[str, BaseTransformation]) -> tuple[AnnData, gpd.GeoDataFrame]:
    zarr_output = cwd / "proseg-output.zarr"

    if zarr_output.exists():  # in versions >= 3.0.0
        _sdata = spatialdata.read_zarr(zarr_output)
        adata, geo_df = _sdata.tables["table"], _sdata.shapes["cell_boundaries"]

        del geo_df.attrs["transform"]
    else:
        import gzip

        counts = pd.read_csv(cwd / "expected-counts.csv.gz")

        obs = pd.read_csv(cwd / "cell-metadata.csv.gz")
        obs.index = obs.index.map(str)

        adata = AnnData(counts, obs=obs)

        with gzip.open(cwd / "cell-polygons.geojson.gz", "rb") as f:
            geo_df = gpd.read_file(f)

    geo_df.crs = None
    geo_df = ShapesModel.parse(geo_df, transformations=transformations)

    return adata, geo_df


def _use_zarr_output(proseg_executable_path: str) -> bool:
    import subprocess

    from packaging.version import InvalidVersion, Version

    result = subprocess.run(f"{proseg_executable_path} --version", shell=True, capture_output=True, text=True)

    try:
        return Version(result.stdout.split()[1]) >= Version("3.0.0")
    except InvalidVersion:
        log.warning(
            "Could not parse the version of proseg. Assumes proseg<3.0.0. We may assume fallback to a different version in the future."
        )
        return False



from pathlib import Path

import dask.array as da
import pandas as pd
from dask_image.imread import imread
from spatialdata import SpatialData
from spatialdata.models import Image2DModel, PointsModel

from ...constants import SopaAttrs
from .utils import _default_image_kwargs


def molecular_cartography(
    path: str | Path,
    region: str,
    image_models_kwargs: dict | None = None,
    imread_kwargs: dict | None = None,
) -> SpatialData:
    """Read *Molecular Cartography* data from *Resolve Bioscience* as a `SpatialData` object.

    Args:
        path: Path to the directory containing the `.tiff` images and `_results.txt` files.
        region: Name of the region to read. The region name can be found before the `_results.txt` file, e.g. `A2-1`.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        imread_kwargs: Keyword arguments passed to `dask_image.imread.imread`.

    Returns:
        A `SpatialData` object representing the *Resolve Bioscience* experiment
    """
    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)

    path = Path(path)
    dataset_id = _get_dataset_id(path, region)

    # Read the points
    transcripts = pd.read_csv(path / f"{dataset_id}_results.txt", sep="\t", header=None)
    transcripts.columns = ["x", "y", "z", "target_name", "unnamed"]

    transcripts = PointsModel.parse(transcripts, feature_key="target_name")
    transcripts_name = f"{dataset_id}_points"

    # Read the images
    images_paths = list(path.glob(f"{dataset_id}_*.tiff"))
    c_coords = [image_path.stem.split("_")[-1] for image_path in images_paths]

    image = Image2DModel.parse(
        da.concatenate([imread(image_path, **imread_kwargs) for image_path in images_paths], axis=0),
        dims=("c", "y", "x"),
        c_coords=c_coords,
        rgb=None,
        **image_models_kwargs,
    )
    image_name = f"{dataset_id}_image"

    return SpatialData(
        images={image_name: image},
        points={transcripts_name: transcripts},
        attrs={
            SopaAttrs.CELL_SEGMENTATION: image_name,
            SopaAttrs.TRANSCRIPTS: transcripts_name,
        },
    )


def _get_dataset_id(path: Path, region: str) -> str:
    _dataset_ids = [path.name[:-12] for path in path.glob("*_results.txt")]
    region_to_id = {dataset_id.split("_")[-1]: dataset_id for dataset_id in _dataset_ids}

    assert region in region_to_id, f"Region {region} not found. Must be one of {list(region_to_id.keys())}"

    return region_to_id[region]



import json
import logging
from pathlib import Path

import geopandas as gpd
import pandas as pd
from anndata import AnnData
from spatialdata import SpatialData

from ...constants import ATTRS_KEY, SopaAttrs, SopaKeys
from ...utils import get_boundaries, get_feature_key, get_spatial_element, get_spatial_image, to_intrinsic
from . import write_cell_categories, write_gene_counts, write_image, write_polygons, write_transcripts
from ._constants import FileNames, experiment_dict
from .utils import explorer_file_path, is_valid_explorer_id

log = logging.getLogger(__name__)


def _check_explorer_directory(path: Path):
    assert not path.exists() or path.is_dir(), (
        "A path to an existing file was provided. It should be a path to a directory."
    )
    path.mkdir(parents=True, exist_ok=True)


def _should_save(mode: str | None, character: str):
    if mode is None:
        return True

    assert len(mode) and mode[0] in [
        "-",
        "+",
    ], "Mode should be a string that starts with '+' or '-'"

    return character in mode if mode[0] == "+" else character not in mode


def write(
    path: str,
    sdata: SpatialData,
    table_key: str = SopaKeys.TABLE,
    image_key: str | None = None,
    shapes_key: str | None = None,
    points_key: str | None = None,
    gene_column: str | None = None,
    pixel_size: float = 0.2125,
    layer: str | None = None,
    polygon_max_vertices: int = 13,
    lazy: bool = True,
    ram_threshold_gb: int | None = 4,
    mode: str | None = None,
    raw_data_path: str | None = None,
    save_h5ad: bool = False,
    run_name: str | None = None,
) -> None:
    """
    Transform a SpatialData object into inputs for the Xenium Explorer - it can be [downloaded for free here](https://www.10xgenomics.com/support/software/xenium-explorer).
    After running this function, double-click on the `experiment.xenium` file to open the explorer.

    !!! note "Quick explorer update"
        If you have already run this function but updated/filtered your table and cells,
        you can simply provide `mode="-it"` to prevent from writing the image and transcript files again, since they are the same.

    Note:
        This function will create up to 7 files, depending on the `SpatialData` object and the arguments:

        - `experiment.xenium` contains some experiment metadata. Double-click on this file to open the Xenium Explorer. This file can also be created with [`write_metadata`](./#sopa.io.explorer.write_metadata).

        - `morphology.ome.tif` is the primary image. This file can also be created with [`write_image`](./#sopa.io.explorer.write_image). Add more images with `align`.

        - `analysis.zarr.zip` contains the cells categories (or clusters), i.e. `adata.obs`. This file can also be created with [`write_cell_categories`](./#sopa.io.explorer.write_cell_categories).

        - `cell_feature_matrix.zarr.zip` contains the cell-by-gene counts. This file can also be created with [`write_gene_counts`](./#sopa.io.explorer.write_gene_counts).

        - `cells.zarr.zip` contains the cells polygon boundaries. This file can also be created with [`write_polygons`](./#sopa.io.explorer.write_polygons).

        - `transcripts.zarr.zip` contains transcripts locations. This file can also be created with [`write_transcripts`](./#sopa.io.explorer.write_transcripts).

        - `adata.h5ad` is the `AnnData` object from the `SpatialData`. This is **not** used by the Explorer, but only saved for convenience.

    Args:
        path: Path to the directory where files will be saved.
        sdata: SpatialData object.
        table_key: Name of the table containing the gene counts or intensities (key of `sdata.tables`). By default, uses `sdata["table"]`.
        image_key: Name of the image of interest (key of `sdata.images`). By default, it will be inferred.
        shapes_key: Name of the cell shapes (key of `sdata.shapes`). By default, it will be inferred from the table.
        points_key: Name of the transcripts (key of `sdata.points`). By default, it will be inferred.
        gene_column: Column name of the points dataframe containing the gene names.
        pixel_size: Number of microns in a pixel. Invalid value can lead to inconsistent scales in the Explorer.
        layer: Layer of the AnnData table where the gene counts are saved. If `None`, uses `table.X`.
        polygon_max_vertices: Maximum number of vertices for the cell polygons.
        lazy: If `True`, will not load the full images in memory (except if the image memory is below `ram_threshold_gb`).
        ram_threshold_gb: Threshold (in gigabytes) from which image can be loaded in memory. If `None`, the image is never loaded in memory.
        mode: String that indicates which files should be created. For instance, `"-it"` means everything except **i**mages and **t**ranscripts, while `"+bocm"` means only **b**oundaries/**o**bservations/**c**ounts/**m**etadata (each letter corresponds to one explorer file). By default, keeps everything.
        save_h5ad: Whether to save the adata as h5ad in the explorer directory (for convenience only, since h5ad is faster to open than the original .zarr table)
        run_name: Name of the run displayed in the Xenium Explorer. If `None`, uses the `image_key`.
    """
    path: Path = Path(path)
    _check_explorer_directory(path)

    image_key, _ = get_spatial_image(sdata, key=image_key, return_key=True)
    preserve_ids: bool | None = None

    # try using symlinks to avoid re-generating large files
    raw_data_path = raw_data_path or sdata.attrs.get(SopaAttrs.XENIUM_OUTPUT_PATH)

    ### Saving table / cell categories / gene counts
    if table_key in sdata.tables:
        adata: AnnData = sdata.tables[table_key]

        _shapes_key = adata.uns[ATTRS_KEY]["region"]
        assert shapes_key is None or _shapes_key == shapes_key, (
            f"Got {shapes_key=}, while the table corresponds to the shapes {_shapes_key}"
        )
        shapes_key = _shapes_key[0] if isinstance(_shapes_key, list) else _shapes_key

        geo_df = sdata[shapes_key]

        preserve_ids = _update_preserve_ids(adata.obs_names, geo_df.index)

        if _should_save(mode, "c"):
            write_gene_counts(path, adata, layer=layer, preserve_ids=preserve_ids)
        if _should_save(mode, "o"):
            write_cell_categories(path, adata)
        if save_h5ad:
            adata.write_h5ad(path / FileNames.H5AD)

    ### Saving cell boundaries
    if not _should_save(mode, "b") and not _should_save(mode, "m"):
        shapes_key, geo_df = None, None
    elif shapes_key is None:
        shapes_key, geo_df = get_boundaries(sdata, return_key=True, warn=True)
    else:
        geo_df = sdata[shapes_key]

    if _should_save(mode, "b") and geo_df is not None:
        geo_df = to_intrinsic(sdata, geo_df, image_key)

        if table_key in sdata.tables:
            geo_df = geo_df.loc[adata.obs[adata.uns[ATTRS_KEY]["instance_key"]]]

        if preserve_ids is None:
            preserve_ids = _update_preserve_ids(geo_df.index)

        write_polygons(path, geo_df, polygon_max_vertices, pixel_size=pixel_size, preserve_ids=preserve_ids)

    ### Saving transcripts
    df = None
    if len(sdata.points):
        df = get_spatial_element(sdata.points, key=points_key or sdata.attrs.get(SopaAttrs.TRANSCRIPTS))

    if _should_save(mode, "t") and not _use_symlink(path, raw_data_path, "transcripts*") and df is not None:
        gene_column = gene_column or get_feature_key(df)
        if gene_column is not None:
            df = to_intrinsic(sdata, df, image_key)
            write_transcripts(path, df, gene_column, pixel_size=pixel_size)
        else:
            log.warning("The argument 'gene_column' has to be provided to save the transcripts")

    ### Saving image
    if _should_save(mode, "i") and not _use_symlink(path, raw_data_path, "morphology*"):
        write_image(
            path,
            sdata[image_key],
            lazy=lazy,
            ram_threshold_gb=ram_threshold_gb,
            pixel_size=pixel_size,
        )

    ### Saving experiment.xenium file
    if _should_save(mode, "m"):
        region_name = _get_region_name(sdata, shapes_key)
        write_metadata(path, run_name or image_key, region_name, _get_n_obs(sdata, geo_df, table_key), pixel_size)

    log.info(f"Saved files in the following directory: {path}")
    log.info(f"You can open the experiment with 'open {path / FileNames.METADATA}'")


def _get_region_name(sdata: SpatialData, shapes_key: str) -> str:
    if SopaAttrs.XENIUM_OUTPUT_PATH not in sdata.attrs:
        return shapes_key
    try:
        with open(Path(sdata.attrs[SopaAttrs.XENIUM_OUTPUT_PATH]) / FileNames.METADATA) as f:
            metadata: dict = json.load(f)
            return f"{metadata['region_name']}-{shapes_key}"  # recover original region name
    except Exception:
        return shapes_key


def _use_symlink(dst: Path, raw_data_path: str | None, pattern: str) -> bool:
    """Try using the Xenium output files when existing to avoid re-generating large files."""
    if raw_data_path is None:
        return False

    sources = list(Path(raw_data_path).glob(pattern))

    for src in sources:
        target = dst / src.name

        if src.is_dir():
            target.mkdir(parents=True, exist_ok=True)

            for sub_src in src.iterdir():
                if not sub_src.is_file():
                    continue

                sub_target = target / sub_src.name

                if sub_target.exists():
                    if not sub_target.is_symlink():
                        return False
                    sub_target.unlink()

                sub_target.symlink_to(sub_src.resolve())
                log.info(f"Created symlink {sub_target} -> {sub_src.resolve()}")
        else:
            if target.exists():
                if not target.is_symlink():  # avoid removing non-symlink files
                    return False
                target.unlink()

            target.symlink_to(src.resolve())
            log.info(f"Created symlink {target} -> {src.resolve()}")

    return len(sources) > 0


def _get_n_obs(sdata: SpatialData, geo_df: gpd.GeoDataFrame, table_key: str) -> int:
    if table_key in sdata.tables:
        return sdata.tables[table_key].n_obs
    return len(geo_df) if geo_df is not None else 0


def write_metadata(
    path: str,
    run_name: str = "NA",
    region_name: str = "NA",
    n_obs: int = 0,
    is_dir: bool = True,
    pixel_size: float = 0.2125,
):
    """Create an `experiment.xenium` file that can be open by the Xenium Explorer.

    Note:
        This function alone is not enough to actually open an experiment. You will need at least to wrun `write_image`, or create all the outputs with `write`.

    Args:
        path: Path to the Xenium Explorer directory where the metadata file will be written
        run_name: Key of `SpatialData` object containing the primary image used on the explorer.
        region_name: Name of the region to be displayed on the Xenium Explorer.
        n_obs: Number of cells
        is_dir: If `False`, then `path` is a path to a single file, not to the Xenium Explorer directory.
        pixel_size: Number of microns in a pixel. Invalid value can lead to inconsistent scales in the Explorer.
    """
    path = Path(path)

    additional_images = {}
    if is_dir:
        mip_file: Path = path / "morphology_mip.ome.tif"

        if mip_file.exists():
            additional_images["morphology_mip_filepath"] = mip_file.name

        focus_file1: Path = path / "morphology_focus.ome.tif"
        focus_file2: Path = path / "morphology_focus" / "morphology_focus_0000.ome.tif"
        focus_file3: Path = path / "morphology_focus" / "ch0000_dapi.ome.tif"
        if focus_file1.exists():
            additional_images["morphology_focus_filepath"] = focus_file1.name
        elif focus_file2.exists():
            additional_images["morphology_focus_filepath"] = "morphology_focus/morphology_focus_0000.ome.tif"
        elif focus_file3.exists():
            additional_images["morphology_focus_filepath"] = "morphology_focus/ch0000_dapi.ome.tif"

    path = explorer_file_path(path, FileNames.METADATA, is_dir)

    metadata = experiment_dict(run_name, region_name, n_obs, pixel_size)
    metadata["images"] |= additional_images

    with open(path, "w") as f:
        json.dump(metadata, f, indent=4)


def _update_preserve_ids(*indices: pd.Index) -> bool:
    if all(index.map(is_valid_explorer_id).all() for index in indices):
        return True

    log.warning(
        "The cell IDs in the table or shapes are not valid Xenium Explorer IDs. "
        "They will be replaced by integers starting from 0 in the output files."
    )
    return False



import typer

from .utils import SDATA_HELPER

app_explorer = typer.Typer()


@app_explorer.command()
def write(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    output_path: str = typer.Option(
        None,
        help="Path to a directory where Xenium Explorer's outputs will be saved. By default, writes to the same path as `sdata_path` but with the `.explorer` suffix",
    ),
    gene_column: str = typer.Option(None, help="Column name of the points dataframe containing the gene names"),
    table_key: str = typer.Option(
        "table",
        help="Sdata key for the table. By default, uses 'table'.",
    ),
    shapes_key: str = typer.Option(
        None,
        help="Sdata key for the boundaries. By default, uses the baysor boundaires, else the cellpose boundaries",
    ),
    pixel_size: float = typer.Option(
        0.2125,
        help="Number of microns in a pixel. Invalid value can lead to inconsistent scales in the Explorer.",
    ),
    lazy: bool = typer.Option(
        True,
        help="If `True`, will not load the full images in memory (except if the image memory is below `ram_threshold_gb`)",
    ),
    ram_threshold_gb: int = typer.Option(
        4,
        help="Threshold (in gygabytes) from which image can be loaded in memory. If `None`, the image is never loaded in memory",
    ),
    mode: str = typer.Option(
        None,
        help="String that indicated which files should be created. `'-ib'` means everything except images and boundaries, while `'+tocm'` means only transcripts/observations/counts/metadata (each letter corresponds to one explorer file). By default, keeps everything",
    ),
    raw_data_path: str = typer.Option(
        None,
        help="Optional path to the original Xenium output directory (to avoid re-generating large files such as transcripts or images). If `None`, tries to find the path from `sdata.attrs`.",
    ),
    save_h5ad: bool = typer.Option(
        True,
        help="Whether to save the adata as h5ad in the explorer directory (for convenience only, since h5ad is faster to open than the original .zarr table)",
    ),
    run_name: str = typer.Option(
        None, help="Name of the run displayed in the Xenium Explorer. If `None`, uses the `image_key`"
    ),
):
    """Convert a spatialdata object to Xenium Explorer's inputs"""
    from pathlib import Path

    from sopa.io.explorer import write
    from sopa.io.standardize import read_zarr_standardized

    sdata = read_zarr_standardized(sdata_path)

    if output_path is None:
        output_path = Path(sdata_path).with_suffix(".explorer")

    write(
        output_path,
        sdata,
        shapes_key=shapes_key,
        table_key=table_key,
        gene_column=gene_column,
        pixel_size=pixel_size,
        lazy=lazy,
        ram_threshold_gb=ram_threshold_gb,
        mode=mode,
        raw_data_path=raw_data_path,
        save_h5ad=save_h5ad,
        run_name=run_name,
    )


@app_explorer.command()
def update_obs(
    adata_path: str = typer.Argument(
        help="Path to the anndata file (`zarr` or `h5ad`) containing the new observations"
    ),
    output_path: str = typer.Argument(
        help="Path to the Xenium Explorer directory (it will update `analysis.zarr.zip`)",
    ),
):
    """Update the cell categories for the Xenium Explorer's (i.e. what's in `adata.obs`). This is useful when you perform analysis and update your `AnnData` object

    Usage:
        Make sure you have already run `sopa explorer write` before. This command should only be used if you updated `adata.obs`
    """
    from pathlib import Path

    import anndata

    from sopa.io.explorer import write_cell_categories

    path = Path(adata_path)

    adata = anndata.io.read_zarr(path) if path.is_dir() else anndata.io.read_h5ad(path)

    write_cell_categories(output_path, adata)


@app_explorer.command()
def add_aligned(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    image_path: str = typer.Argument(
        help="Path to the image file to be added (`.ome.tif` used in the explorer during alignment)"
    ),
    transformation_matrix_path: str = typer.Argument(
        help="Path to the `matrix.csv` file returned by the Explorer after alignment"
    ),
    image_key: str = typer.Option(
        None,
        help="Optional original-image key (of sdata.images) on which the new image will be aligned. This doesn't need to be provided if there is only one image, or a cell-segmentation image.",
    ),
    overwrite: bool = typer.Option(False, help="Whether to overwrite the image if existing"),
):
    """After alignment on the Xenium Explorer, add an image to the SpatialData object"""
    import spatialdata

    from sopa import io
    from sopa.io.explorer.images import align

    sdata = spatialdata.read_zarr(sdata_path)
    image = io.ome_tif(image_path, as_image=True)

    align(sdata, image, transformation_matrix_path, overwrite=overwrite, image_key=image_key)



import logging

import geopandas as gpd
import numpy as np
import pandas as pd
from anndata import AnnData
from scipy.sparse import csr_matrix
from spatialdata import SpatialData
from spatialdata.models import TableModel

from ..constants import ATTRS_KEY, SopaAttrs, SopaKeys
from ..io.explorer.utils import str_cell_id
from ..utils import add_spatial_element, get_boundaries, get_spatial_element, get_spatial_image, validated_channel_names
from . import aggregate_bins, count_transcripts
from . import aggregate_channels as _aggregate_channels

log = logging.getLogger(__name__)


def aggregate(
    sdata: SpatialData,
    aggregate_genes: bool | None = None,
    aggregate_channels: bool = True,
    image_key: str | None = None,
    points_key: str | None = None,
    gene_column: str | None = None,
    shapes_key: str | None = None,
    bins_key: str | None = None,
    expand_radius_ratio: float | None = None,
    min_transcripts: int = 0,
    min_intensity_ratio: float = 0.1,
    no_overlap: bool = False,
    key_added: str | None = "table",
):
    """Aggregate gene counts and/or channel intensities over a `SpatialData` object to create an `AnnData` table (saved in `sdata["table"]`).

    !!! info
        The main arguments are `sdata`, `aggregate_genes`, and `aggregate_channels`. The rest of the arguments are optional and will be inferred from the data if not provided.

        - If channels are aggregated and not genes, then `sdata['table'].X` will contain the mean channel intensities per cell.
        - If genes are aggregated and not channels, then `sdata['table'].X` will contain the gene counts per cell.
        - If both genes and channels are aggregated, then `sdata['table'].X` will contain the gene counts per cell and `sdata['table'].obsm['intensities']` will contain the mean channel intensities per cell.

    Args:
        sdata: A `SpatialData` object
        aggregate_genes: Whether to aggregate gene counts. If None, it will be inferred from the data.
        aggregate_channels: Whether to aggregate channel intensities inside cells.
        image_key: Key of `sdata` with the image channels to be averaged. By default, uses the segmentation image.
        points_key: Key of `sdata` with the points dataframe representing the transcripts. Inferred by default.
        gene_column: Key of `sdata[points_key]` with the gene names. Inferred by default.
        shapes_key: Key of `sdata` with the shapes corresponding to the cells boundaries. Inferred by default.
        bins_key: Key of `sdata` with the table corresponding to the bin-by-gene table of gene counts (e.g., for Visium HD data). Inferred by default.
        expand_radius_ratio: Ratio to expand the cells polygons for channels averaging. For instance, a ratio of 0.5 expands the shape radius by 50%. If `None` (default), use 1 if we aggregate bins data, and 0 otherwise.
        min_transcripts: Min number of transcripts to keep a cell.
        min_intensity_ratio: Min ratio of the 90th quantile of the mean channel intensity to keep a cell.
        no_overlap: *Experimental feature*: If `True`, the (expanded) cells will not overlap for channels and bins aggregation.
        key_added: Key to save the table in `sdata.tables`. If `None`, it will be `f"{shapes_key}_table"`.
    """
    assert points_key is None or bins_key is None, "Provide either `points_key` or `bins_key`, not both."

    if points_key is None:
        bins_key = bins_key or sdata.attrs.get(SopaAttrs.BINS_TABLE)

    if (bins_key is None) and (aggregate_genes or (aggregate_genes is None and sdata.points)):
        assert sdata.points, (
            "No points in the SpatialData object. You must have points, or set the `bins_key` argument (for VisiumHD-like data)."
        )

        points_key, _ = get_spatial_element(
            sdata.points, key=points_key or sdata.attrs.get(SopaAttrs.TRANSCRIPTS), return_key=True
        )

    aggr = Aggregator(sdata, image_key=image_key, shapes_key=shapes_key, bins_key=bins_key, points_key=points_key)

    if key_added is None:
        key_added = f"{aggr.shapes_key}_{SopaKeys.TABLE}"
        log.info(f"key_added is None, saving the table as '{key_added}' by default.")

    aggr.compute_table(
        aggregate_genes=aggregate_genes,
        aggregate_channels=aggregate_channels,
        expand_radius_ratio=expand_radius_ratio,
        min_transcripts=min_transcripts,
        min_intensity_ratio=min_intensity_ratio,
        gene_column=gene_column,
        no_overlap=no_overlap,
        key_added=key_added,
    )


class Aggregator:
    """Perform transcript count and channel averaging over a `SpatialData` object"""

    table: AnnData | None

    def __init__(
        self,
        sdata: SpatialData,
        image_key: str | None = None,
        shapes_key: str | None = None,
        bins_key: str | None = None,
        points_key: str | None = None,
    ):
        """
        Args:
            sdata: A `SpatialData` object
            image_key: Key of `sdata` with the image to be averaged. If only one image, this does not have to be provided
            shapes_key: Key of `sdata` with the shapes corresponding to the cells boundaries
            bins_key: Key of `sdata` with the table corresponding to the bins table of gene counts (e.g., for Visium HD data)
            points_key: Key of `sdata` with the points dataframe representing the transcripts
        """
        self.sdata = sdata
        self.bins_key = bins_key
        self.points_key = points_key
        self.shapes_key, self.geo_df = get_boundaries(sdata, return_key=True, key=shapes_key)
        self.table = None

        if not sdata.images:
            self.image_key, self.image = "None", None
        else:
            self.image_key, self.image = get_spatial_image(sdata, image_key, return_key=True)

    def already_has_valid_table(self, key_added: str) -> bool:
        if (key_added not in self.sdata.tables) or (self.sdata[key_added].n_obs != len(self.geo_df)):
            return False

        log.info("Found existing table, transcripts are not count again.")
        self.table = self.sdata[key_added]

        return True

    def filter_cells(self, where_filter: np.ndarray, reason: str | None = None):
        log.info(f"Filtering {where_filter.sum()} cells" + (f" due to {reason}" if reason else ""))

        self.geo_df = self.geo_df[~where_filter]
        self.sdata.shapes[self.shapes_key] = self.geo_df

        if self.table is not None:
            self.table = self.table[~where_filter]

    def compute_table(
        self,
        aggregate_genes: bool | None = None,
        aggregate_channels: bool = True,
        gene_column: str | None = None,
        expand_radius_ratio: float | None = None,
        min_transcripts: int = 0,
        min_intensity_ratio: float = 0,
        no_overlap: bool = False,
        key_added: str = SopaKeys.TABLE,
    ):
        if aggregate_genes is None:
            aggregate_genes = (
                (self.table is not None and isinstance(self.table.X, csr_matrix))
                or gene_column is not None
                or self.bins_key is not None
                or self.points_key is not None
            )

        if aggregate_channels and self.image is None:
            log.warning("No image found to aggregate channels. Use `aggregate_channels=False`")
            aggregate_channels = False

        assert aggregate_genes or aggregate_channels, (
            "At least one of `aggregate_genes` or `aggregate_channels` must be True"
        )

        if aggregate_genes:
            if self.bins_key is not None:
                self.table = aggregate_bins(
                    self.sdata,
                    self.shapes_key,
                    self.bins_key,
                    expand_radius_ratio=1 if expand_radius_ratio is None else expand_radius_ratio,
                    no_overlap=no_overlap,
                )
            elif not self.already_has_valid_table(key_added):
                self.table = count_transcripts(
                    self.sdata, gene_column, shapes_key=self.shapes_key, points_key=self.points_key
                )

            if min_transcripts > 0:
                self.filter_cells(self.table.X.sum(axis=1) < min_transcripts, f"transcript count < {min_transcripts}")

        if aggregate_channels:
            mean_intensities = _aggregate_channels(
                self.sdata,
                image_key=self.image_key,
                shapes_key=self.shapes_key,
                expand_radius_ratio=expand_radius_ratio or 0,
                no_overlap=no_overlap,
            )

            if min_intensity_ratio > 0:
                means = mean_intensities.mean(axis=1)
                intensity_threshold = min_intensity_ratio * np.quantile(means, 0.9)
                where_filter = means < intensity_threshold
                self.filter_cells(where_filter, f"mean channel intensity < {intensity_threshold:.2f}")
                mean_intensities = mean_intensities[~where_filter]

            if aggregate_genes:
                self.table.obsm[SopaKeys.INTENSITIES_OBSM] = pd.DataFrame(
                    mean_intensities,
                    columns=validated_channel_names(self.image),
                    index=self.table.obs_names,
                )
            else:
                self.table = AnnData(
                    mean_intensities,
                    var=pd.DataFrame(index=validated_channel_names(self.image)),
                    obs=pd.DataFrame(index=self.geo_df.index),
                )

        self.sdata.shapes[self.shapes_key] = self.geo_df

        self.table.uns[SopaKeys.UNS_KEY] = {
            SopaKeys.UNS_HAS_TRANSCRIPTS: aggregate_genes,
            SopaKeys.UNS_HAS_INTENSITIES: aggregate_channels,
        }

        self.add_standardized_table(key_added)

    def add_standardized_table(self, key_added: str):
        add_standardized_table(
            self.sdata, self.table, self.geo_df, self.shapes_key, key_added, image_key=self.image_key
        )


def add_standardized_table(
    sdata: SpatialData,
    table: AnnData,
    geo_df: gpd.GeoDataFrame,
    shapes_key: str,
    table_key: str,
    image_key: str | None = None,
):
    table.obs_names = list(map(str_cell_id, range(table.n_obs)))
    geo_df.index = list(table.obs_names)

    add_spatial_element(sdata, shapes_key, geo_df)

    table.obsm["spatial"] = np.array([[centroid.x, centroid.y] for centroid in geo_df.centroid])
    table.obs[SopaKeys.REGION_KEY] = pd.Series(shapes_key, index=table.obs_names, dtype="category")
    table.obs[SopaKeys.SLIDE_KEY] = pd.Series(image_key or "None", index=table.obs_names, dtype="category")
    table.obs[SopaKeys.INSTANCE_KEY] = geo_df.index
    table.obs[SopaKeys.AREA_OBS] = geo_df.area.values

    if ATTRS_KEY in table.uns:
        del table.uns[ATTRS_KEY]

    table = TableModel.parse(
        table,
        region_key=SopaKeys.REGION_KEY,
        region=shapes_key,
        instance_key=SopaKeys.INSTANCE_KEY,
    )

    add_spatial_element(sdata, table_key, table)



from .patches import Patches2D
from .loader import TileLoader
from .factory import make_image_patches, make_transcript_patches
from .cluster import cluster_embeddings
from .infer import compute_embeddings



import logging
from functools import partial

import dask.dataframe as dd
import geopandas as gpd
import numpy as np
import pandas as pd
from anndata import AnnData
from dask.diagnostics import ProgressBar
from scipy.sparse import csr_matrix
from spatialdata import SpatialData

from .. import settings
from ..constants import SopaAttrs, SopaKeys
from ..utils import get_boundaries, get_feature_key, get_spatial_element, to_intrinsic

log = logging.getLogger(__name__)


def count_transcripts(
    sdata: SpatialData,
    gene_column: str | None = None,
    shapes_key: str | None = None,
    points_key: str | None = None,
    geo_df: gpd.GeoDataFrame | None = None,
    only_excluded: bool = False,
) -> AnnData:
    """Counts transcripts per cell.

    Args:
        sdata: A `SpatialData` object
        gene_column: Column of the transcript dataframe containing the gene names
        shapes_key: Key of `sdata` containing the cell boundaries. If only one `shapes` element, this does not have to be provided.
        points_key: Key of `sdata` containing the transcripts. If only one `points` element, this does not have to be provided.
        geo_df: If the cell boundaries are not yet in `sdata`, a `GeoDataFrame` can be directly provided for cell boundaries
        only_excluded: By default, the genes matching the pattern in `sopa.settings.gene_exclude_pattern` are excluded from the count. If `only_excluded=True`, it counts **only** these excluded genes.

    Returns:
        An `AnnData` object of shape `(n_cells, n_genes)` with the counts per cell
    """
    points_key, points = get_spatial_element(
        sdata.points, key=points_key or sdata.attrs.get(SopaAttrs.TRANSCRIPTS), return_key=True
    )

    if geo_df is None:
        geo_df = get_boundaries(sdata, key=shapes_key)
        geo_df = to_intrinsic(sdata, geo_df, points_key)

    gene_column = gene_column or get_feature_key(points, raise_error=True)

    log.info(f"Aggregating transcripts over {len(geo_df)} cells")
    return _count_transcripts_aligned(geo_df, points, gene_column, only_excluded)


def _count_transcripts_aligned(
    geo_df: gpd.GeoDataFrame,
    points: dd.DataFrame,
    value_key: str,
    only_excluded: bool = False,
) -> AnnData:
    """Count transcripts per cell. The cells and points have to be aligned (i.e., in the same coordinate system)

    Args:
        geo_df: Cells geometries
        points: Transcripts dataframe
        value_key: Key of `points` containing the genes names
        only_excluded: Whether to count the valid genes or the excluded ones

    Returns:
        An `AnnData` object of shape `(n_cells, n_genes)` with the counts per cell
    """
    points[value_key] = points[value_key].astype("category").cat.as_known()
    gene_names = points[value_key].cat.categories.astype(str)

    X = csr_matrix((len(geo_df), len(gene_names)), dtype=int)
    adata = AnnData(X=X, var=pd.DataFrame(index=gene_names))
    adata.obs_names = geo_df.index.astype(str)

    geo_df = geo_df.reset_index()

    X_partitions = []

    with ProgressBar():
        points.map_partitions(
            partial(
                _add_csr,
                X_partitions,
                geo_df,
                gene_column=value_key,
                gene_names=gene_names,
                only_excluded=only_excluded,
            ),
            meta=(),
        ).compute()

    for X_partition in X_partitions:
        adata.X += X_partition

    if settings.gene_exclude_pattern is not None:
        matching = adata.var_names.str.match(settings.gene_exclude_pattern, case=False, na=True)
        adata = (adata[:, matching] if only_excluded else adata[:, ~matching]).copy()

    return adata


def _add_csr(
    X_partitions: list[csr_matrix],
    geo_df: gpd.GeoDataFrame,
    partition: pd.DataFrame,
    gene_column: str,
    gene_names: list[str],
    only_excluded: bool,
) -> None:
    if settings.gene_exclude_pattern is not None:
        matching = partition[gene_column].str.match(settings.gene_exclude_pattern, case=False, na=True)
        partition = partition[matching] if only_excluded else partition[~matching]

    if SopaKeys.LOW_QUALITY_TRANSCRIPT_KEY in partition.columns:
        partition = partition[~partition[SopaKeys.LOW_QUALITY_TRANSCRIPT_KEY]]

    points_gdf = gpd.GeoDataFrame(partition, geometry=gpd.points_from_xy(partition["x"], partition["y"]))
    joined = geo_df.sjoin(points_gdf)
    cells_indices, column_indices = joined.index, joined[gene_column].cat.codes

    cells_indices = cells_indices[column_indices >= 0]
    column_indices = column_indices[column_indices >= 0]

    X_partition = csr_matrix(
        (np.full(len(cells_indices), 1), (cells_indices, column_indices)),
        shape=(len(geo_df), len(gene_names)),
    )

    X_partitions.append(X_partition)



import os
import sys
import warnings
from functools import partial
from typing import Callable

import numpy as np
from spatialdata import SpatialData

from ...constants import SopaKeys
from ._custom import custom_staining_based


def stardist(
    sdata: SpatialData,
    model_type: str = "2D_versatile_he",
    image_key: str | None = None,
    channels: list[str] | str | None = None,
    min_area: int = 0,
    delete_cache: bool = True,
    recover: bool = False,
    prob_thresh: float = 0.2,
    nms_thresh: float = 0.6,
    clip_limit: float = 0,
    clahe_kernel_size: int | list[int] | None = None,
    gaussian_sigma: float = 0,
    key_added: str = SopaKeys.STARDIST_BOUNDARIES,
    **stardist_eval_kwargs: int,
):
    """Run [Stardist](https://github.com/stardist/stardist) segmentation on a SpatialData object, and add a GeoDataFrame containing the cell boundaries.

    !!! warning "Stardist installation"
        Make sure to install the stardist extra (`pip install 'sopa[stardist]'`) for this method to work.

    Args:
        sdata: A `SpatialData` object
        model_type: Stardist model name.
        image_key: Name of the image in `sdata` to be used for segmentation.
        channels: One or a list of channel names used for segmentation. None assumes RGB image.
        min_area: Minimum area of a cell to be considered.
        delete_cache: Whether to delete the cache after segmentation.
        recover: If `True`, recover the cache from a failed segmentation, and continue.
        prob_thresh: Stardist `prob_thresh` parameter.
        nms_thresh: Stardist `nms_thresh` parameter.
        clip_limit: Parameter for skimage.exposure.equalize_adapthist (applied before running stardist)
        clahe_kernel_size: Parameter for skimage.exposure.equalize_adapthist (applied before running stardist)
        gaussian_sigma: Parameter for scipy gaussian_filter (applied before running stardist)
        key_added: Name of the shapes element to be added to `sdata`.
        **stardist_eval_kwargs: Kwargs to be provided to `model.predict_instances` (where `model` is a `stardist.models.StarDist2D` object)
    """
    method = stardist_patch(
        model_type=model_type,
        prob_thresh=prob_thresh,
        nms_thresh=nms_thresh,
        **stardist_eval_kwargs,
    )

    custom_staining_based(
        sdata,
        method,
        channels=channels,
        image_key=image_key,
        min_area=min_area,
        delete_cache=delete_cache,
        recover=recover,
        clip_limit=clip_limit,
        clahe_kernel_size=clahe_kernel_size,
        gaussian_sigma=gaussian_sigma,
        cache_dir_name=key_added,
        key_added=key_added,
    )


def stardist_patch(
    model_type: str = "2D_versatile_he",
    prob_thresh: float = 0.2,
    nms_thresh: float = 0.6,
    channels: list[str] | str | None = None,  # for the CLI to work, as "channels" will be provided
    **stardist_eval_kwargs: int,
) -> Callable:
    try:
        from csbdeep.utils import normalize
        from stardist.models import StarDist2D
    except ImportError:
        raise ImportError("To use stardist, you need its corresponding sopa extra: `pip install 'sopa[stardist]'`.")

    def _(
        patch: np.ndarray,
        model_type: str,
        prob_thresh: float,
        nms_thresh: float,
        **stardist_eval_kwargs,
    ):
        with SuppressPrintsAndWarnings():
            model = StarDist2D.from_pretrained(model_type)

            patch = normalize(patch.transpose(1, 2, 0), clip=True)
            mask, _ = model.predict_instances(
                patch, prob_thresh=prob_thresh, nms_thresh=nms_thresh, **stardist_eval_kwargs
            )

            return mask

    return partial(
        _,
        model_type=model_type,
        prob_thresh=prob_thresh,
        nms_thresh=nms_thresh,
        **stardist_eval_kwargs,
    )


class SuppressPrintsAndWarnings:
    def __enter__(self):
        # Suppress stdout (print statements)
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")
        # Suppress warnings
        self._original_warnings_filter = warnings.filters[:]
        warnings.simplefilter("ignore")

    def __exit__(self, *args, **kwargs):
        # Restore stdout
        sys.stdout.close()
        sys.stdout = self._original_stdout
        # Restore warnings filter
        warnings.filters = self._original_warnings_filter



import logging

import dask
import geopandas as gpd
import numpy as np
import numpy.ma as ma
import shapely
from dask.diagnostics import ProgressBar
from shapely.geometry import Polygon, box
from spatialdata import SpatialData
from xarray import DataArray

from ..shapes import expand_radius, pixel_outer_bounds, rasterize
from ..utils import get_boundaries, get_spatial_image, to_intrinsic

log = logging.getLogger(__name__)

AVAILABLE_MODES = ["average", "min", "max"]


def aggregate_channels(
    sdata: SpatialData,
    image_key: str | None = None,
    shapes_key: str | None = None,
    expand_radius_ratio: float = 0,
    mode: str = "average",
    no_overlap: bool = False,
) -> np.ndarray:
    """Aggregate the channel intensities per cell (either `"average"`, or take the `"min"` / `"max"`).

    Args:
        sdata: A `SpatialData` object
        image_key: Key of `sdata` containing the image. If only one `images` element, this does not have to be provided.
        shapes_key: Key of `sdata` containing the cell boundaries. If only one `shapes` element, this does not have to be provided.
        expand_radius_ratio: Cells polygons will be expanded by `expand_radius_ratio * mean_radius`. This help better aggregate boundary stainings.
        mode: Aggregation mode. One of `"average"`, `"min"`, `"max"`. By default, average the intensity inside the cell mask.
        no_overlap: If `True`, the (expanded) cells will not overlap.

    Returns:
        A numpy `ndarray` of shape `(n_cells, n_channels)`
    """
    assert mode in AVAILABLE_MODES, f"Invalid {mode=}. Available modes are {AVAILABLE_MODES}"

    image = get_spatial_image(sdata, image_key)

    geo_df = get_boundaries(sdata, key=shapes_key)
    geo_df = to_intrinsic(sdata, geo_df, image)
    geo_df = expand_radius(geo_df, expand_radius_ratio, no_overlap=no_overlap)

    return _aggregate_channels_aligned(image, geo_df, mode)


def _aggregate_channels_aligned(image: DataArray, geo_df: gpd.GeoDataFrame | list[Polygon], mode: str) -> np.ndarray:
    """Average channel intensities per cell. The image and cells have to be aligned, i.e. be on the same coordinate system.

    Args:
        image: A `DataArray` of shape `(n_channels, y, x)`
        geo_df: A `GeoDataFrame` whose geometries are cell boundaries (polygons)

    Returns:
        A numpy `ndarray` of shape `(n_cells, n_channels)`
    """
    log.info(f"Aggregating channels intensity over {len(geo_df)} cells with {mode=}")

    cells = geo_df if isinstance(geo_df, list) else list(geo_df.geometry)
    tree = shapely.STRtree(cells)

    n_channels = len(image.coords["c"])
    areas = np.zeros(len(cells))
    if mode == "min":
        aggregation = np.full((len(cells), n_channels), fill_value=np.inf)
    else:
        aggregation = np.zeros((len(cells), n_channels))

    chunk_sizes = image.data.chunks
    offsets_y = np.cumsum(np.pad(chunk_sizes[1], (1, 0), "constant"))
    offsets_x = np.cumsum(np.pad(chunk_sizes[2], (1, 0), "constant"))

    def _average_chunk_inside_cells(chunk, iy, ix):
        ymin, ymax = offsets_y[iy], offsets_y[iy + 1]
        xmin, xmax = offsets_x[ix], offsets_x[ix + 1]

        patch = box(xmin, ymin, xmax, ymax)
        intersections = tree.query(patch, predicate="intersects")

        for index in intersections:
            cell = cells[index]
            bounds = pixel_outer_bounds(cell.bounds)

            sub_image = chunk[
                :,
                max(bounds[1] - ymin, 0) : bounds[3] - ymin,
                max(bounds[0] - xmin, 0) : bounds[2] - xmin,
            ]

            if sub_image.shape[1] == 0 or sub_image.shape[2] == 0:
                continue

            mask = rasterize(cell, sub_image.shape[1:], bounds)

            areas[index] += np.sum(mask)

            if mode == "min":
                masked_image = ma.masked_array(sub_image, 1 - np.repeat(mask[None], n_channels, axis=0))
                aggregation[index] = np.minimum(aggregation[index], masked_image.min(axis=(1, 2)))
            elif mode in ["average", "max"]:
                func = np.sum if mode == "average" else np.max
                values = func(sub_image * mask, axis=(1, 2))

                if mode == "average":
                    aggregation[index] += values
                else:
                    aggregation[index] = np.maximum(aggregation[index], values)

    with ProgressBar():
        tasks = [
            dask.delayed(_average_chunk_inside_cells)(chunk, iy, ix)
            for iy, row in enumerate(image.chunk({"c": -1}).data.to_delayed()[0])
            for ix, chunk in enumerate(row)
        ]
        dask.compute(tasks)

    if mode == "average":
        return aggregation / areas[:, None].clip(1)
    else:
        return aggregation



import logging
from math import ceil
from pathlib import Path

import geopandas as gpd
import numpy as np
import zarr
from shapely.geometry import MultiPolygon, Polygon

from ._constants import FileNames, cell_summary_attrs, group_attrs
from .utils import explorer_file_path, int_cell_id

log = logging.getLogger(__name__)

TOLERANCE_STEP = 0.5


def pad_polygon(polygon: Polygon, max_vertices: int, tolerance: float = TOLERANCE_STEP) -> np.ndarray:
    """Transform the polygon to have the desired number of vertices

    Args:
        polygon: A `shapely` polygon
        max_vertices: The desired number of vertices
        tolerance: The step of tolerance used for simplification. At each step, we increase the tolerance of this value until the polygon is simplified enough.

    Returns:
        A 2D array representing the polygon vertices
    """
    n_vertices = len(polygon.exterior.coords)
    assert n_vertices >= 3

    if tolerance > 50:
        log.warning(
            f"Polygon with {n_vertices} vertices could not be simplified to {max_vertices} vertices. Using a circle instead."
        )
        polygon = polygon.centroid.buffer(np.sqrt(polygon.area) / 2)
        return pad_polygon(polygon, max_vertices, TOLERANCE_STEP)

    coords = polygon.exterior.coords._coords

    if n_vertices == max_vertices:
        return coords.flatten()

    if n_vertices < max_vertices:
        return np.pad(coords, ((0, max_vertices - n_vertices), (0, 0)), mode="edge").flatten()

    polygon = polygon.simplify(tolerance=tolerance)
    return pad_polygon(polygon, max_vertices, tolerance + TOLERANCE_STEP)


def write_polygons(
    path: Path,
    geo_df: gpd.GeoDataFrame,
    max_vertices: int,
    is_dir: bool = True,
    pixel_size: float = 0.2125,
    preserve_ids: bool = True,
) -> None:
    """Write a `cells.zarr.zip` file containing the cell polygonal boundaries

    Args:
        path: Path to the Xenium Explorer directory where the transcript file will be written
        geo_df: A GeoDataFrame containing the `shapely` polygons to be written
        max_vertices: The number of vertices per polygon (they will be transformed to have the right number of vertices)
        is_dir: If `False`, then `path` is a path to a single file, not to the Xenium Explorer directory.
        pixel_size: Number of microns in a pixel. Invalid value can lead to inconsistent scales in the Explorer.
        preserve_ids: If `True`, will preserve the cell IDs from `geo_df.index` in the explorer (they need to be valid Xenium Explorer IDs). If `False`, will use new IDs for the output file.
    """
    assert isinstance(geo_df, gpd.GeoDataFrame), "geo_df must be a GeoDataFrame"

    polygons = geo_df.geometry.values

    assert all(isinstance(geom, (Polygon, MultiPolygon)) for geom in polygons), (
        "All geometries must be a `shapely.Polygon` or `shapely.MultiPolygon`"
    )

    if any(isinstance(geom, MultiPolygon) for geom in polygons):
        log.warning("Some cells are MultiPolygons. Only the polygon with the largest area will be kept for each cell.")
        polygons = [_to_largest_polygon(p) for p in polygons]

    path = explorer_file_path(path, FileNames.SHAPES, is_dir)

    log.info(f"Writing {len(polygons)} cell polygons")
    coordinates = np.stack([pad_polygon(p, max_vertices) for p in polygons])
    coordinates *= pixel_size

    num_cells = len(coordinates)
    cells_fourth = ceil(num_cells / 4)
    cells_half = ceil(num_cells / 2)

    GROUP_ATTRS = group_attrs()
    GROUP_ATTRS["number_cells"] = num_cells

    polygon_vertices = np.stack([coordinates, coordinates])
    num_points = polygon_vertices.shape[2]
    n_vertices = num_points // 2

    with zarr.ZipStore(path, mode="w") as store:
        g = zarr.group(store=store)
        g.attrs.put(GROUP_ATTRS)

        g.array(
            "polygon_vertices",
            polygon_vertices,
            dtype="float32",
            chunks=(1, cells_fourth, ceil(num_points / 4)),
        )

        cell_id = np.ones((num_cells, 2))
        cell_id[:, 0] = int_cell_id(geo_df.index) if preserve_ids else np.arange(num_cells)
        g.array("cell_id", cell_id, dtype="uint32", chunks=(cells_half, 1))

        cell_summary = np.zeros((num_cells, 7))
        cell_summary[:, 2] = [p.area for p in polygons]
        g.array(
            "cell_summary",
            cell_summary,
            dtype="float64",
            chunks=(num_cells, 1),
        )
        g["cell_summary"].attrs.put(cell_summary_attrs())

        g.array(
            "polygon_num_vertices",
            np.full((2, num_cells), n_vertices),
            dtype="int32",
            chunks=(1, cells_half),
        )

        g.array(
            "seg_mask_value",
            np.arange(num_cells),
            dtype="uint32",
            chunks=(cells_half,),
        )


def _to_largest_polygon(polygon: Polygon | MultiPolygon) -> Polygon:
    if isinstance(polygon, Polygon):
        return polygon
    return max(polygon.geoms, key=lambda x: x.area)



import logging
from pathlib import Path

from spatialdata import SpatialData

from ...constants import ATTRS_KEY, SopaAttrs, SopaKeys
from ...utils import ensure_string_channel_names
from .utils import _default_image_kwargs

log = logging.getLogger(__name__)


def xenium(
    path: str | Path,
    image_models_kwargs: dict | None = None,
    imread_kwargs: dict | None = None,
    cells_boundaries: int = False,
    cells_table: int = False,
    nucleus_labels: int = False,
    cells_labels: int = False,
    cells_as_circles: int = False,
    nucleus_boundaries: int = False,
    qv_threshold: int | None = None,
    **kwargs: int,
) -> SpatialData:
    """Read Xenium data as a `SpatialData` object. For more information, refer to [spatialdata-io](https://spatialdata.scverse.org/projects/io/en/latest/generated/spatialdata_io.xenium.html).

    This function reads the following files:
        - `transcripts.parquet`: transcripts locations and names
        - `experiment.xenium`: metadata file
        - `morphology_focus.ome.tif`: morphology image (or a directory, for recent versions of the Xenium)


    Args:
        path: Path to the Xenium directory containing all the experiment files
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        imread_kwargs: Keyword arguments passed to `dask_image.imread.imread`.
        cells_boundaries: Whether to read cell boundaries
        cells_table: Whether to read cell table
        nucleus_labels: Whether to read nucleus labels
        cells_labels: Whether to read cell labels
        cells_as_circles: Whether to read cells as circles
        nucleus_boundaries: Whether to read nucleus boundaries
        qv_threshold: Whether to add a "low_quality_transcript" column to transcripts. Transcripts with a QV value below `qv_threshold` will not be used during segmentation.
        kwargs: Additional keyword arguments passed to `spatialdata_io.xenium

    Returns:
        A `SpatialData` object representing the Xenium experiment
    """
    from spatialdata_io.readers.xenium import xenium as xenium_spatialdata_io

    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)

    sdata: SpatialData = xenium_spatialdata_io(
        path,
        cells_table=cells_table,
        nucleus_labels=nucleus_labels,
        cells_labels=cells_labels,
        cells_as_circles=cells_as_circles,
        nucleus_boundaries=nucleus_boundaries,
        cells_boundaries=cells_boundaries,
        image_models_kwargs=image_models_kwargs,
        imread_kwargs=imread_kwargs,
        **kwargs,
    )

    if "table" in sdata.tables:
        sdata["table"].uns[ATTRS_KEY]["region"] = "cell_boundaries"
        sdata["table"].obs["region"] = "cell_boundaries"
        sdata["table"].obs["region"] = sdata["table"].obs["region"].astype("category")

    ensure_string_channel_names(sdata)

    ### Add Sopa attributes to detect the spatial elements
    if "morphology_focus" in sdata.images:
        sdata.attrs[SopaAttrs.CELL_SEGMENTATION] = "morphology_focus"

    if "he_image" in sdata.images:
        sdata.attrs[SopaAttrs.TISSUE_SEGMENTATION] = "he_image"

    if "transcripts" in sdata.points:
        sdata.attrs[SopaAttrs.TRANSCRIPTS] = "transcripts"

        if "cell_id" in sdata.points["transcripts"].columns:
            sdata.attrs[SopaAttrs.PRIOR_TUPLE_KEY] = ["cell_id", "UNASSIGNED"]

        if qv_threshold is not None:
            assert "qv" in sdata.points["transcripts"].columns, "QV column not found in `sdata['transcripts']`"

            sdata.points["transcripts"][SopaKeys.LOW_QUALITY_TRANSCRIPT_KEY] = (
                sdata.points["transcripts"]["qv"] < qv_threshold
            )

    sdata.attrs[SopaAttrs.XENIUM_OUTPUT_PATH] = str(Path(path).resolve())

    return sdata



def _check_zip(names: list[str]):
    for name in names:
        if isinstance(name, str):
            assert name.endswith(".zip"), "Intermediate files names must end with .zip"


def _default_boundary_dir(sdata_path: str, directory_name: str):
    from pathlib import Path

    from sopa.constants import SopaFiles

    temp_dir = Path(sdata_path) / SopaFiles.SOPA_CACHE_DIR / directory_name
    temp_dir.mkdir(parents=True, exist_ok=True)

    return temp_dir


def _log_whether_to_resolve(patch_index: int | None, delete_cache: bool = True):
    import logging

    log = logging.getLogger(__name__)

    if patch_index is None and delete_cache:
        log.info("Segmentation is already resolved. Don't run `sopa resolve`.")


SDATA_HELPER = "Path to the SpatialData `.zarr` directory"



import logging

log = logging.getLogger(__name__)


class ColorFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    blue = "\x1b[36;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    prefix = "[%(levelname)s] (%(name)s)"
    suffix = "%(message)s"

    FORMATS = {
        logging.DEBUG: f"{grey}{prefix}{reset} {suffix}",
        logging.INFO: f"{blue}{prefix}{reset} {suffix}",
        logging.WARNING: f"{yellow}{prefix}{reset} {suffix}",
        logging.ERROR: f"{red}{prefix}{reset} {suffix}",
        logging.CRITICAL: f"{bold_red}{prefix}{reset} {suffix}",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def configure_logger(log: logging.Logger):
    log.setLevel(logging.INFO)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(ColorFormatter())

    log.addHandler(consoleHandler)
    log.propagate = False



import logging
from typing import Callable

import tqdm
from anndata import AnnData
from spatialdata import SpatialData
from spatialdata.models import TableModel

from ..constants import SopaKeys
from ..utils import add_spatial_element
from . import TileLoader

log = logging.getLogger(__name__)


def compute_embeddings(
    sdata: SpatialData,
    model: Callable | str,
    patch_width: int,
    patch_overlap: int = 0,
    level: int | None = 0,
    magnification: int | None = None,
    image_key: str | None = None,
    batch_size: int = 32,
    device: str | None = None,
    data_parallel: bool | list[int] = False,
    roi_key: str | None = SopaKeys.ROI,
    key_added: str | None = None,
    **kwargs: int,
) -> str:
    """It creates patches, runs a computer vision model on each patch, and store the embeddings of each all patches as an [`AnnData` object](https://anndata.readthedocs.io/en/stable/). This is mostly useful for WSI images.

    !!! info
        The `AnnData` object will be saved into the `SpatialData` object with the key `"{model_name}_embeddings"` (see the `model_name` argument below), except if `key_added` is provided.
        The shapes of the patches will be saved with the key `"embeddings_patches"`.

    !!! warning
        In addition to the WSI extra (`pip install 'sopa[wsi]'`) and depending on the model used, you might need to install additional dependencies. Also, CONCH requires to be logged in Hugging Face and having approved their License.

    Args:
        sdata: A `SpatialData` object
        model: A supported model name (`resnet50`, `histo_ssl`, `dinov2`, `hoptimus0`, or `conch`), or a callable that takes as an input a tensor of size `(batch_size, channels, x, y)` and returns a vector for each tile `(batch_size, emb_dim)`.
        patch_width: Width of the patches in pixels.
        patch_overlap: Width of the overlap between the patches in pixels.
        level: Image level on which the processing is performed. Either `level` or `magnification` should be provided.
        magnification: The target magnification on which the processing is performed. If `magnification` is provided, the `level` argument will be automatically computed.
        image_key: Optional image key of the image. By default, uses the only image (if only one) or the image used for cell or tissue segmentation.
        batch_size: Mini-batch size used during inference.
        device: Device used for the computer vision model.
        data_parallel: If `True`, the model will be run in data parallel mode. If a list of GPUs is provided, the model will be run in data parallel mode on the specified GPUs.
        roi_key: Optional name of the shapes that needs to touch the patches. Patches that do not touch any shape will be ignored. If `None`, all patches will be used. By default, uses the tissue segmentation if available.
        key_added: Optional name of the spatial element that will be added (storing the embeddings).
        **kwargs: Additional keyword arguments passed to the `Patches2D` constructor.

    Returns:
        The name of the `AnnData` table that was added to the `SpatialData` object.
    """
    try:
        import torch
    except ImportError:
        raise ImportError(
            "For patch embedding, you need `torch` (and perhaps `torchvision`). Consider installing the sopa WSI extra: `pip install 'sopa[wsi]'`."
        )
    from . import models

    if isinstance(model, str):
        assert model in models.available_models, (
            f"'{model}' is not a valid model name. Valid names are: {', '.join(list(models.available_models.keys()))}"
        )
        model_name, model = model, models.available_models[model]()
    else:
        model_name = model.__class__.__name__

    if device is not None:
        model.to(device)

    if data_parallel:
        ids = data_parallel if isinstance(data_parallel, list) else list(range(torch.cuda.device_count()))
        model = torch.nn.DataParallel(model, device_ids=ids)

    tile_loader = TileLoader(sdata, patch_width, image_key, level, magnification, patch_overlap, roi_key)

    log.info(f"Processing {len(tile_loader)} patches extracted from level {tile_loader.level}")

    predictions = []
    with torch.no_grad():
        for i in tqdm.tqdm(range(0, len(tile_loader), batch_size)):
            batch = tile_loader[i : i + batch_size]
            embedding: torch.Tensor = model(batch.to(device))
            assert len(embedding.shape) == 2, "The model must have the signature (B, C, Y, X) -> (B, C)"

            predictions.append(embedding.cpu())

    predictions = torch.cat(predictions)
    if len(predictions.shape) == 1:
        predictions = torch.unsqueeze(predictions, 1)

    patches = tile_loader.patches
    patches.add_shapes(key_added=SopaKeys.EMBEDDINGS_PATCHES)

    gdf = sdata[SopaKeys.EMBEDDINGS_PATCHES]

    adata = AnnData(predictions.numpy())
    adata.obs["region"] = SopaKeys.EMBEDDINGS_PATCHES
    adata.obs["instance"] = gdf.index.values
    adata = TableModel.parse(
        adata,
        region=SopaKeys.EMBEDDINGS_PATCHES,
        region_key="region",
        instance_key="instance",
    )
    adata.obsm["spatial"] = patches.centroids()
    adata.uns["embedding_config"] = {
        "patch_width": patch_width,
        "patch_overlap": patch_overlap,
        "magnification": magnification,
        "level": tile_loader.level,
        "level_downsample": tile_loader.level_downsample,
        "tile_resize_factor": tile_loader.tile_resize_factor,
        "model_name": model_name,
    }

    key_added = key_added or f"{model_name}_embeddings"
    add_spatial_element(sdata, key_added, adata)

    return key_added



from torch import nn


class DummyFeatures(nn.Module):
    def __init__(self):
        super().__init__()

    def __call__(self, x):
        """(B, C, Y, X) -> (B, C)"""
        return x[:, :, 0, 0]



from .explorer import write
from .reader.cosmx import cosmx
from .reader.merscope import merscope
from .reader.xenium import xenium
from .reader.macsima import macsima
from .reader.phenocycler import phenocycler
from .reader.hyperion import hyperion
from .reader.utils import ome_tif
from .reader.wsi import wsi, wsi_autoscale
from .reader.generic import aicsimageio, bioio
from .reader.visium_hd import visium_hd
from .reader.molecular_cartography import molecular_cartography
from .report import write_report
from ..utils.data import blobs, toy_dataset
from . import explorer, standardize



import logging

import dask
import numpy as np
from spatialdata import SpatialData
from xarray import DataArray, DataTree

from .. import settings
from ..constants import SopaAttrs, SopaKeys
from ..io.reader._wsi_reader import get_reader
from ..utils import get_spatial_element
from .patches import Patches2D

log = logging.getLogger(__name__)


class TileLoader:
    def __init__(
        self,
        sdata: SpatialData,
        patch_width: int,
        image_key: str | None = None,
        level: int | None = 0,
        magnification: int | None = None,
        patch_overlap: int = 0,
        roi_key: str | None = SopaKeys.ROI,
    ):
        multiscale_image = _get_image_for_inference(sdata, image_key)
        self.image, self.level, self.tile_resize_factor, self.level_downsample = _get_extraction_parameters(
            multiscale_image, level, magnification
        )

        _backend = multiscale_image.attrs.get("backend")
        _path = multiscale_image.attrs.get("path")

        try:
            if settings.native_read_region and _backend is not None:
                self.slide = get_reader(_backend)(_path)
            else:
                self.slide = get_reader("xarray")(multiscale_image)
        except Exception as e:
            log.warning(
                f"Exception raised for '{_backend}' and path '{_path}'. Falling back to xarray reader. Error: {e}"
            )
            self.slide = get_reader("xarray")(multiscale_image)

        self.patch_width = int(patch_width / self.tile_resize_factor)
        self.resized_patch_width = patch_width
        self.patches = Patches2D(sdata, self.image, self.patch_width, patch_overlap=patch_overlap, roi_key=roi_key)

    def _numpy_patch(self, box: tuple[int, int, int, int]) -> np.ndarray:
        """
        Extract a numpy patch from the image given a bounding box
        and pads a patch to a specific width since some patches might be smaller (e.g., on edges)
        """
        image_patch = np.array(
            self.slide.read_region(
                (int(box[0] * self.level_downsample), int(box[1] * self.level_downsample)),
                self.level,
                (box[2] - box[0], box[3] - box[1]),
            )
        )

        pad_x, pad_y = self.patch_width - image_patch.shape[0], self.patch_width - image_patch.shape[1]
        padded_patch = np.pad(image_patch, ((0, pad_x), (0, pad_y), (0, 0)))
        return padded_patch

    def get_batch(self, bboxes: np.ndarray):
        """Retrives a batch of patches using the bboxes"""
        import torch

        delayed_patches = [dask.delayed(self._numpy_patch)(box) for box in bboxes]
        batch = np.array(dask.compute(*delayed_patches))

        batch = torch.tensor(batch, dtype=torch.float32) / 255.0
        batch = batch.movedim(-1, 1)

        if self.tile_resize_factor != 1:
            from torchvision.transforms import Resize

            dim = (self.resized_patch_width, self.resized_patch_width)
            batch = Resize(dim)(batch)

        assert len(batch.shape) == 4

        return batch

    def get_PIL_patch(self, idx: int):
        """Retrives a PIL patch using the index"""
        from PIL import Image

        return Image.fromarray(np.array(self[idx][0].movedim(0, -1) * 255, dtype="uint8"))

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx: int | slice):
        bboxes = self.patches.bboxes[idx]
        batch = self.get_batch(np.atleast_2d(bboxes))
        return batch


def _get_image_for_inference(sdata: SpatialData, image_key: str | None = None) -> DataArray | DataTree:
    if image_key is not None:
        return get_spatial_element(sdata.images, key=image_key)

    cell_image = sdata.attrs.get(SopaAttrs.CELL_SEGMENTATION)
    tissue_image = sdata.attrs.get(SopaAttrs.TISSUE_SEGMENTATION)

    assert cell_image is None or tissue_image is None or cell_image == tissue_image, (
        "When different images are existing for cell and tissue segmentation, you need to provide the `image_key` argument"
    )

    return get_spatial_element(sdata.images, key=cell_image or tissue_image)


def _get_extraction_parameters(
    image: DataArray | DataTree, level: int | None, magnification: int | None
) -> tuple[DataArray, int, float]:
    if isinstance(image, DataArray):
        assert level == 0, "Level must be 0 when using a DataArray"
        assert magnification is None, "Magnification must be None when the image is not multi scale"
        return image, 0, 1, 1

    if level < 0:
        assert isinstance(level, int) and level >= -len(image.keys()), "Invalid level"
        level = len(image.keys()) + level

    if magnification is None:
        tile_resize_factor, level_downsample = 1, 1
        if level is None:
            log.warning("Both level and magnification arguments are None. Using level=0 by default.")
            level = 0
    else:
        level, tile_resize_factor, level_downsample = _get_level_for_magnification(image, magnification)

    assert f"scale{level}" in image, f"Level {level} not found in the image scales"

    image = next(iter(image[f"scale{level}"].values()))

    return image, level, tile_resize_factor, level_downsample


def _get_level_for_magnification(image: DataArray | DataTree, magnification: int) -> int:
    """Return the best level for a given downsampling factor"""
    slide_metadata, backend = image.attrs.get("metadata", {}), image.attrs.get("backend")

    assert slide_metadata, "No `metadata` field found in the image attributes"
    assert backend is not None, "No backend found in the image metadata, can not infer the level for the magnification"
    assert "level_downsamples" in slide_metadata, "Missing `level_downsamples` in the image metadata"
    assert "properties" in slide_metadata, "Missing `properties` in the image metadata"

    objective_power = slide_metadata["properties"].get(f"{backend}.objective-power")
    mpp_x = slide_metadata["properties"].get(f"{backend}.mpp-x")

    if objective_power:
        downsample = int(objective_power) / magnification
    elif mpp_x:
        mpp_objective = min([80, 40, 20, 10, 5], key=lambda obj: abs(10 / obj - float(mpp_x)))
        downsample = mpp_objective / magnification
    else:
        raise ValueError("No objective-power or mpp-x information found in the metadata")

    if downsample < 1.0:
        log.warning(
            f"The requested magnification {magnification}x is higher than the objective power {objective_power}x. "
            f"Using the highest available magnification with upscaling."
        )

    level = _get_best_level_for_downsample(slide_metadata["level_downsamples"], downsample)
    level_downsample = slide_metadata["level_downsamples"][level]
    tile_resize_factor = level_downsample / downsample

    return level, tile_resize_factor, level_downsample


def _get_best_level_for_downsample(level_downsamples: list[float], downsample: float) -> int:
    """Return the best level for a given downsampling factor"""
    if downsample <= 1.0:
        return 0
    for level, ds in enumerate(level_downsamples):
        if ds > downsample:
            return level - 1
    return len(level_downsamples) - 1



import logging

import geopandas as gpd
import numpy as np
from anndata import AnnData
from scipy.sparse import coo_matrix, csr_matrix
from spatialdata import SpatialData
from tqdm import tqdm

from ..shapes import expand_radius
from ..utils import to_intrinsic

log = logging.getLogger(__name__)


def aggregate_bins(
    sdata: SpatialData,
    shapes_key: str,
    bins_key: str,
    expand_radius_ratio: float = 0,
    no_overlap: bool = False,
) -> AnnData:
    """Aggregate bins (for instance, from Visium HD data) into cells.

    Args:
        sdata: The `SpatialData` object
        shapes_key: Key of the shapes containing the cell boundaries
        bins_key: Key of the table containing the bin-by-gene counts
        expand_radius_ratio: Cells polygons will be expanded by `expand_radius_ratio * mean_radius`. This help better aggregate bins from the cytoplasm.
        no_overlap: If `True`, bins belonging to multiple cells will be assigned to only one, based on transcript-profile proximity.

    Returns:
        An `AnnData` object of shape with the cell-by-gene count matrix
    """
    bins_table: AnnData = sdata.tables[bins_key]

    bins_shapes_key = sdata.get_annotated_regions(bins_table)
    bins_shapes_key = bins_shapes_key[0] if isinstance(bins_shapes_key, list) else bins_shapes_key
    bins = sdata.shapes[bins_shapes_key].loc[sdata.get_instance_key_column(bins_table).values]
    bins = gpd.GeoDataFrame(geometry=bins.centroid.values)  # bins as points

    cells = to_intrinsic(sdata, shapes_key, bins_shapes_key).reset_index(drop=True)
    cells = expand_radius(cells, expand_radius_ratio, no_overlap=False)

    bin_within_cell = gpd.sjoin(bins, cells)

    indices_matrix = csr_matrix(
        (np.full(len(bin_within_cell), 1), (bin_within_cell["index_right"], bin_within_cell.index)),
        shape=(len(cells), len(bins)),
    )

    if no_overlap:
        log.warning("Unique bin assignments is currently experimental. Any feedback on GitHub is welcome.")
        indices_matrix = _get_unique_bins_assignments(indices_matrix, bins_table)

    adata = AnnData(indices_matrix @ bins_table.X, obs=cells[[]], var=bins_table.var)
    adata.obsm["spatial"] = np.stack([cells.centroid.x, cells.centroid.y], axis=1)
    adata.obsm["bins_assignments"] = indices_matrix
    return adata


def _get_unique_bins_assignments(
    indices_matrix: csr_matrix,
    bins_table: AnnData,
    n_components: int = 50,
) -> csr_matrix:
    shared_bins: np.ndarray = (indices_matrix.sum(0) >= 2).A1

    if not shared_bins.any():
        return indices_matrix

    indices_matrix: coo_matrix = indices_matrix.tocoo()
    no_overlap: coo_matrix = indices_matrix.copy()
    no_overlap.data[shared_bins[no_overlap.col]] = 0
    no_overlap.eliminate_zeros()
    no_overlap = no_overlap.tolil()

    adata_unique_map = AnnData(no_overlap @ bins_table.X)

    X_unique, X_shared = _pca_representation(n_components, adata_unique_map.X, bins_table.X[shared_bins])

    for bin_index_among_shared, bin_index in enumerate(tqdm(np.where(shared_bins)[0])):
        cell_indices = indices_matrix.row[indices_matrix.col == bin_index]

        distances: np.ndarray = ((X_unique[cell_indices] - X_shared[bin_index_among_shared]) ** 2).sum(1)
        cell_index = cell_indices[distances.argmin()]

        no_overlap[cell_index, bin_index] = 1

    return no_overlap.tocsr()


def _pca_representation(
    n_components: int, adata_unique_map: np.ndarray | csr_matrix, shared_bins: np.ndarray | csr_matrix
) -> tuple[np.ndarray, np.ndarray]:
    from sklearn.decomposition import PCA

    n_components = min(n_components, shared_bins.shape[1] - 1, shared_bins.shape[0] - 1, adata_unique_map.shape[0] - 1)
    pca = PCA(n_components=n_components)

    X_unique = pca.fit_transform(_log1p_tpm(adata_unique_map))
    X_shared = pca.transform(_log1p_tpm(shared_bins))

    return X_unique, X_shared


def _log1p_tpm(x: np.ndarray | csr_matrix) -> np.ndarray:
    if isinstance(x, np.ndarray):
        x = x / (x.sum(axis=1, keepdims=True) + 1e-8) * 1e6
    else:
        x = x / (x.sum(axis=1) + 1e-8) * 1e6
    return np.log1p(x)



import logging
from functools import partial
from pathlib import Path

from spatialdata import SpatialData

from ... import settings
from ...constants import SopaFiles, SopaKeys
from ...utils import (
    delete_transcripts_patches_dirs,
    get_feature_key,
    get_transcripts_patches_dirs,
    set_boundaries_attrs,
)
from .utils import _check_transcript_patches, _get_executable_path, resolve

log = logging.getLogger(__name__)


def baysor(
    sdata: SpatialData,
    config: dict | str | None = None,
    min_area: int = 0,
    delete_cache: bool = True,
    recover: bool = False,
    force: bool = False,
    scale: float | None = None,
    key_added: str = SopaKeys.BAYSOR_BOUNDARIES,
    patch_index: int | None = None,
):
    """Run [Baysor](https://kharchenkolab.github.io/Baysor/dev/) segmentation on a SpatialData object, and add a GeoDataFrame containing the cell boundaries.

    !!! warning "Baysor installation"
        Make sure to install [Baysor](https://kharchenkolab.github.io/Baysor/dev/installation/), and either have the executable at `~/.julia/bin/baysor`, or create an alias called
        `baysor` that points to the binary executable. Also, you'll need to install
        sopa with the baysor extra: `pip install 'sopa[baysor]'` (basically, this installs `toml` and `loompy`).

    !!! info "Inferred config"
        If the `config` argument is not provided, the configuration is inferred.
        If [sopa.make_transcript_patches][] was run with a `prior_shapes_key`, the configuration is inferred based on the prior segmentation.
        Otherwise, the configuration is inferred based on the `scale` parameter (you'll need to provide it).

    Args:
        sdata: A `SpatialData` object.
        config: Optional configuration dictionary or path to a TOML file containing a valid Baysor config. By default, a configuration is inferred based on the cell area of the prior segmentation, or based on the `scale` parameter.
        min_area: Minimal area (in microns^2) of a cell to be considered.
        delete_cache: Whether to delete the cache after segmentation.
        recover: If `True`, recover the cache from a failed segmentation, and continue.
        force: If `True`, ignore failed patches and continue with the successful ones.
        scale: The typical cell radius in microns. If `config` is not provided, the configuration is inferred based on this parameter. Else, it will overwrite the existing scale in the config.
        key_added: Name of the shapes element to be added to `sdata.shapes`.
        patch_index: Index of the patch to segment (we do not recommend to set this argument). By default, segment all patches.
    """
    _check_transcript_patches(sdata)

    prior_shapes_key = (
        SopaKeys.SOPA_PRIOR if SopaKeys.PRIOR_SHAPES_KEY in sdata.shapes[SopaKeys.TRANSCRIPTS_PATCHES] else None
    )

    if config is None or not len(config):
        config = _get_default_config(sdata, prior_shapes_key, scale)
    elif scale is not None:
        assert "segmentation" in config, "The provided config should contain a 'segmentation' key"
        config["segmentation"]["scale"] = scale

    assert "data" in config, "The provided config should contain a 'data' key"

    if "gene" not in config["data"]:
        config["data"]["gene"] = str(get_feature_key(sdata, raise_error=True))

    if "scale" in config["segmentation"] and config["segmentation"]["scale"] >= 20:
        log.warning(
            f"The provided scale ({config['segmentation']['scale']}) seems large. Make sure the scale is in microns, and not in pixels. If you are using CosMX data, note that the transcripts coordinates are now in microns."
        )

    baysor_command = _get_baysor_command(prior_shapes_key)

    baysor_patch = BaysorPatch(baysor_command, config, force=force, capture_output=patch_index is None)

    if patch_index is not None:
        patch_dir = get_transcripts_patches_dirs(sdata, patch_index)
        baysor_patch(patch_dir)
        return

    patches_dirs = get_transcripts_patches_dirs(sdata)

    remaining_patches_dirs = (
        [patch_dir for patch_dir in patches_dirs if not (patch_dir / "segmentation_counts.loom").exists()]
        if recover
        else patches_dirs
    )

    settings._run_with_backend([partial(baysor_patch, patch_dir) for patch_dir in remaining_patches_dirs])

    if force:
        patches_dirs = [patch_dir for patch_dir in patches_dirs if (patch_dir / "segmentation_counts.loom").exists()]
        assert patches_dirs, "Baysor failed on all patches"

    resolve(sdata, patches_dirs, gene_column=config["data"]["gene"], min_area=min_area, key_added=key_added)

    set_boundaries_attrs(sdata, key_added)

    if delete_cache:
        delete_transcripts_patches_dirs(sdata)


class BaysorPatch:
    def __init__(
        self,
        baysor_command: str,
        config: dict | str,
        force: bool = False,
        capture_output: bool = True,
    ):
        self.baysor_command = baysor_command
        self.config = config
        self.force = force
        self.capture_output = capture_output

    def __call__(self, patch_dir: Path):
        _copy_segmentation_config(patch_dir / SopaFiles.TOML_CONFIG_FILE, self.config)

        import subprocess

        result = subprocess.run(self.baysor_command, cwd=patch_dir, shell=True, capture_output=self.capture_output)

        if result.returncode != 0 or not (patch_dir / "segmentation_counts.loom").exists():
            if self.force:
                log.warning(f"Baysor error on patch {patch_dir.resolve()} with command `{self.baysor_command}`")
                return
            raise subprocess.CalledProcessError(
                returncode=result.returncode,
                cmd=self.baysor_command,
                output=result.stdout,
                stderr=result.stderr,
            )


def _get_baysor_command(prior_shapes_key: str | None) -> str:
    baysor_executable_path = _get_executable_path("baysor", ".julia")

    use_polygons_format_argument = _use_polygons_format_argument(baysor_executable_path)
    polygon_format = (
        "--polygon-format GeometryCollection" if use_polygons_format_argument else "--save-polygons GeoJSON"
    )  # depends on the version of baysor

    prior_suffix = f" :{prior_shapes_key}" if prior_shapes_key else ""  # use a prior segmentation

    return f"{baysor_executable_path} run {polygon_format} -c config.toml transcripts.csv" + prior_suffix


def _use_polygons_format_argument(baysor_executable_path: str) -> bool:
    import subprocess

    from packaging.version import InvalidVersion, Version

    result = subprocess.run(f"{baysor_executable_path} run --version", shell=True, capture_output=True, text=True)

    try:
        return Version(result.stdout) >= Version("0.7.0")
    except InvalidVersion:
        log.warning("Could not parse the version of baysor. Assumes baysor >= 0.7.0.")
        return True


def _get_default_config(sdata: SpatialData, prior_shapes_key: str | None, scale: float | None) -> dict:
    assert prior_shapes_key is not None or scale is not None, (
        "The config can't be inferred. Choose among the following solutions:\n"
        "   - Provide the `scale` argument (typical cell radius in microns)\n"
        "   - Run `sopa.make_transcript_patches(...)` with `prior_shapes_key`\n"
        "   - Provide the `config` argument, containing a valid Baysor config."
    )

    config = {
        "data": {
            "x": "x",
            "y": "y",
            "gene": str(get_feature_key(sdata, raise_error=True)),
            "min_molecules_per_gene": 10,
            "min_molecules_per_cell": 20,
            "force_2d": True,
        },
        "segmentation": {"prior_segmentation_confidence": 0.8},
    }

    if scale is not None:
        config["segmentation"]["scale"] = scale

    log.info(f"The Baysor config was not provided, using the following by default:\n{config}")

    return config


def _copy_segmentation_config(path: Path | str, config: dict | str):
    """Copy the segmentation config to a file (`.json` or `.toml`).

    Args:
        path: Where the config will be saved
        config: Dictionnary config, or path to an existing config file (json or toml)
    """
    path = Path(path)

    if isinstance(config, str):
        import shutil

        shutil.copy(config, path)
        return

    assert path.suffix == ".toml"

    try:
        import toml
    except ImportError:
        raise ImportError(
            "To use baysor, you need its corresponding sopa extra: `pip install 'sopa[baysor]'`.\
            \nAlso, make sure to install the baysor executable (https://github.com/kharchenkolab/Baysor)."
        )

    with open(path, "w") as f:
        toml.dump(config, f)
        return



class FileNames:
    IMAGE = "morphology.ome.tif"
    POINTS = "transcripts.zarr.zip"
    SHAPES = "cells.zarr.zip"
    H5AD = "adata.h5ad"
    TABLE = "cell_feature_matrix.zarr.zip"
    CELL_CATEGORIES = "analysis.zarr.zip"
    METADATA = "experiment.xenium"


class ExplorerConstants:
    GRID_SIZE = 250
    TILE_SIZE = 1024
    QUALITY_SCORE = 40
    MICRONS_TO_PIXELS = 4.705882
    PIXELS_TO_MICRONS = 0.2125

    COLORS = ["white", 400, 500, 600, 700]
    NUCLEUS_COLOR = "white"
    KNOWN_CHANNELS = {"DAPI": "white", "DNA1": "white", "DNA2": "white", "DAPI (000)": "white"}


class Versions:
    EXPERIMENT = [6, 0, 0]
    GROUPS = [5, 0]
    CELL_CATEGORIES = [1, 0]


def cell_categories_attrs() -> dict:
    return {
        "major_version": Versions.CELL_CATEGORIES[0],
        "minor_version": Versions.CELL_CATEGORIES[1],
        "number_groupings": 0,
        "grouping_names": [],
        "group_names": [],
    }


def cell_summary_attrs() -> dict:
    return {
        "column_descriptions": [
            "Cell centroid in X",
            "Cell centroid in Y",
            "Cell area",
            "Nucleus centroid in X",
            "Nucleus centroid in Y",
            "Nucleus area",
            "z_level",
        ],
        "column_names": [
            "cell_centroid_x",
            "cell_centroid_y",
            "cell_area",
            "nucleus_centroid_x",
            "nucleus_centroid_y",
            "nucleus_area",
            "z_level",
        ],
    }


def group_attrs() -> dict:
    return {
        "major_version": Versions.GROUPS[0],
        "minor_version": Versions.GROUPS[1],
        "name": "CellSegmentationDataset",
        "polygon_set_descriptions": [
            "NA",
            "NA",
        ],
        "polygon_set_display_names": ["Nucleus boundaries", "Cell boundaries"],
        "polygon_set_names": ["nucleus", "cell"],
        "spatial_units": "microns",
    }


def experiment_dict(run_name: str, region_name: str, num_cells: int, pixel_size: float) -> dict:
    return {
        "major_version": Versions.EXPERIMENT[0],
        "minor_version": Versions.EXPERIMENT[1],
        "patch_version": Versions.EXPERIMENT[2],
        "run_name": run_name,
        "region_name": region_name,
        "experiment_uuid": "N/A",
        "panel_tissue_type": "N/A",
        "run_start_time": "N/A",
        "preservation_method": "N/A",
        "num_cells": num_cells,
        "transcripts_per_cell": 0,
        "transcripts_per_100um": 0,
        "cassette_name": "N/A",
        "slide_id": "N/A",
        "panel_design_id": "N/A",
        "panel_name": "N/A",
        "panel_organism": "Human",
        "panel_num_targets_predesigned": 0,
        "panel_num_targets_custom": 0,
        "pixel_size": pixel_size,
        "instrument_sn": "N/A",
        "instrument_sw_version": "N/A",
        "analysis_sw_version": "xenium-1.3.0.5",
        "cassette_uuid": "",
        "roi_uuid": "",
        "z_step_size": 3.0,
        "well_uuid": "",
        "calibration_uuid": "N/A",
        "images": {
            "morphology_filepath": "morphology.ome.tif",
        },
        "xenium_explorer_files": {
            "transcripts_zarr_filepath": "transcripts.zarr.zip",
            "cells_zarr_filepath": "cells.zarr.zip",
            "cell_features_zarr_filepath": "cell_feature_matrix.zarr.zip",
            "analysis_zarr_filepath": "analysis.zarr.zip",
            "analysis_summary_filepath": "analysis_summary.html",
        },
    }


def image_metadata(channel_names: list[str], pixel_size: float) -> dict:
    return {
        "SignificantBits": 8,
        "PhysicalSizeX": pixel_size,
        "PhysicalSizeXUnit": "µm",
        "PhysicalSizeY": pixel_size,
        "PhysicalSizeYUnit": "µm",
        "Channel": {"Name": channel_names},
    }



from typing import Callable, Literal

import scanpy as sc
from anndata import AnnData
from spatialdata import SpatialData


def leiden_clustering(adata: AnnData, flavor: Literal["leidenalg", "igraph"] = "igraph", **kwargs):
    adata = adata.copy()
    sc.pp.pca(adata)
    sc.pp.neighbors(adata)
    sc.tl.leiden(adata, flavor=flavor, **kwargs)
    return adata.obs["leiden"].values


def kmeans_clustering(adata: AnnData, **kwargs):
    from sklearn.cluster import KMeans

    kmeans = KMeans(**kwargs)
    return kmeans.fit_predict(adata.X)


METHODS_DICT = {
    "leiden": leiden_clustering,
    "kmeans": kmeans_clustering,
}


def cluster_embeddings(
    sdata: SpatialData | None,
    element: AnnData | str,
    method: Callable | str = "leiden",
    key_added: str = "cluster",
    **method_kwargs: str,
) -> None:
    """Create clusters of the patches embeddings (obtained from [sopa.patches.compute_embeddings][]).

    Info:
        The clusters are added to the `key_added` column of the "inference_patches" shapes (`key_added='cluster'` by default).

    Args:
        sdata: A `SpatialData` object. Can be `None` if element is an `AnnData` object.
        element: The `AnnData` containing the embeddings, or the name of the element
        method: Callable that takes as an AnnData object and returns an array of clusters of size `n_obs`, or an available method name (`leiden` or `kmeans`)
        key_added: The key containing the clusters to be added to the `element.obs`
        method_kwargs: kwargs provided to the method callable
    """
    if isinstance(element, str):
        element: AnnData = sdata.tables[element]

    if isinstance(method, str):
        assert method in METHODS_DICT, f"Method {method} is not available. Use one of: {', '.join(METHODS_DICT.keys())}"
        method = METHODS_DICT[method]

    element.obs[key_added] = method(element, **method_kwargs)
    element.obs[key_added] = element.obs[key_added].astype("category")



from torch import nn


class ConchFeatures(nn.Module):
    def __init__(self):
        super().__init__()
        self.conch, self.transform = _import_conch()
        self.conch.eval()

    def forward(self, x):
        return self.conch(self.transform(x))


def _import_conch():
    try:
        import einops_exts  # noqa: F401
        import timm  # noqa: F401
        from torchvision import transforms
        from transformers import AutoModel
    except ImportError:
        raise ImportError("To use CONCH, please install the CONCH dependencies: transformers, einops_exts, timm.")

    titan = AutoModel.from_pretrained("MahmoodLab/TITAN", trust_remote_code=True)
    conch, _ = titan.return_conch()
    transform = transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))

    return conch, transform



import logging
import warnings
from functools import partial
from typing import Callable

import numpy as np
from packaging.version import Version
from spatialdata import SpatialData

from ...constants import SopaKeys
from ._custom import custom_staining_based

log = logging.getLogger(__name__)


def cellpose(
    sdata: SpatialData,
    channels: list[str] | str,
    diameter: int,
    model_type: str = "cyto3",
    pretrained_model: str | None = None,
    gpu: bool = False,
    image_key: str | None = None,
    min_area: int | None = None,
    delete_cache: bool = True,
    recover: bool = False,
    flow_threshold: float = 2,
    cellprob_threshold: float = -6,
    clip_limit: float = 0.2,
    clahe_kernel_size: int | list[int] | None = None,
    gaussian_sigma: float = 1,
    key_added: str = SopaKeys.CELLPOSE_BOUNDARIES,
    cellpose_model_kwargs: dict | None = None,
    **cellpose_eval_kwargs: int,
):
    """Run [Cellpose](https://cellpose.readthedocs.io/en/latest/) segmentation on a SpatialData object, and add a GeoDataFrame containing the cell boundaries.

    !!! warning "Cellpose installation"
        Make sure to install the cellpose extra (`pip install 'sopa[cellpose]'`) for this method to work.

    !!! info "Diameter parameter"
        The `diameter` parameter is used to estimate the expected cell diameter (in pixels). This is a crucial parameter for the segmentation.

    Args:
        sdata: A `SpatialData` object
        channels: Name of the channel(s) to be used for segmentation. If one channel, must be a nucleus channel. If a `list` of channels, it must be a cytoplasmic channel and then a nucleus channel.
        diameter: The Cellpose parameter for the expected cell diameter (in pixel).
        model_type: Cellpose model type.
        pretrained_model: Path to the pretrained model to be loaded, or `None`
        gpu: Whether to use GPU for segmentation.
        image_key: Name of the image in `sdata` to be used for segmentation.
        min_area: Minimum area of a cell to be considered. By default, it is calculated based on the `diameter` parameter.
        delete_cache: Whether to delete the cache after segmentation.
        recover: If `True`, recover the cache from a failed segmentation, and continue.
        flow_threshold: Cellpose `flow_threshold` parameter.
        cellprob_threshold: Cellpose `cellprob_threshold` parameter.
        clip_limit: Parameter for skimage.exposure.equalize_adapthist (applied before running cellpose)
        clahe_kernel_size: Parameter for skimage.exposure.equalize_adapthist (applied before running cellpose)
        gaussian_sigma: Parameter for scipy gaussian_filter (applied before running cellpose)
        key_added: Name of the shapes element to be added to `sdata`.
        cellpose_model_kwargs: Dictionary of kwargs to be provided to the `cellpose.models.CellposeModel` object.
        **cellpose_eval_kwargs: Kwargs to be provided to `model.eval` (where `model` is a `cellpose.models.CellposeModel` object)
    """
    channels = channels if isinstance(channels, list) else [channels]

    method = cellpose_patch(
        diameter=diameter,
        channels=channels,
        model_type=model_type,
        pretrained_model=pretrained_model,
        gpu=gpu,
        flow_threshold=flow_threshold,
        cellprob_threshold=cellprob_threshold,
        cellpose_model_kwargs=cellpose_model_kwargs,
        **cellpose_eval_kwargs,
    )

    if min_area is None:
        min_area = (diameter / 2) ** 2  # by default, about 15% of the "normal cell" area

    custom_staining_based(
        sdata,
        method,
        channels,
        image_key=image_key,
        min_area=min_area,
        delete_cache=delete_cache,
        recover=recover,
        clip_limit=clip_limit,
        clahe_kernel_size=clahe_kernel_size,
        gaussian_sigma=gaussian_sigma,
        cache_dir_name=key_added,
        key_added=key_added,
    )


def cellpose_patch(
    diameter: float,
    channels: list[str],
    model_type: str = "cyto3",
    pretrained_model: str | None = None,
    gpu: bool = False,
    cellpose_model_kwargs: dict | None = None,
    **cellpose_eval_kwargs: int,
) -> Callable:
    """Creation of a callable that runs Cellpose segmentation on a patch

    Args:
        diameter: Cellpose diameter parameter
        channels: List of channel names
        model_type: Cellpose model type
        pretrained_model: Path to the pretrained model to be loaded, or `None`
        gpu: Whether to use GPU for segmentation.
        cellpose_model_kwargs: Kwargs to be provided to the `cellpose.models.CellposeModel` object
        **cellpose_eval_kwargs: Kwargs to be provided to `model.eval` (where `model` is a `cellpose.models.CellposeModel` object)

    Returns:
        A `callable` whose input is an image of shape `(C, Y, X)` and output is a cell mask of shape `(Y, X)`. Each mask value `>0` represent a unique cell ID
    """
    try:
        from cellpose import models, version
    except ImportError:
        raise ImportError("To use cellpose, you need its corresponding sopa extra: `pip install 'sopa[cellpose]'`.")

    if Version(version) >= Version("4.0.0"):
        if not pretrained_model:
            log.info(f"You use cellpose={version}, which requires a `pretrained_model`. Defaulting to 'cpsam'.")
            pretrained_model = "cpsam"

        if not gpu:
            log.warning(
                f"You use cellpose={version}, which can be slow without a GPU. Consider using `gpu=True`, or downgrading to `cellpose<4.0.0`."
            )

    def _(
        patch: np.ndarray,
        diameter: float,
        channels: list[str],
        model_type: str,
        pretrained_model: str | None = None,
        cellpose_model_kwargs: dict | None = None,
        **cellpose_eval_kwargs: int,
    ):
        warnings.filterwarnings("ignore", message="You are using `torch.load` with `weights_only=False`")

        cellpose_model_kwargs = cellpose_model_kwargs or {}

        if pretrained_model:
            model = models.CellposeModel(gpu=gpu, pretrained_model=pretrained_model, **cellpose_model_kwargs)
        else:
            model = models.Cellpose(gpu=gpu, model_type=model_type, **cellpose_model_kwargs)

        if isinstance(channels, str) or len(channels) == 1:
            channels = [0, 0]  # gray scale
        elif len(channels) == 2:
            channels = [1, 2]
        else:
            raise ValueError(f"Provide 1 or 2 channels. Found {len(channels)}")

        mask, *_ = model.eval(patch, diameter=diameter, channels=channels, **cellpose_eval_kwargs)
        return mask

    return partial(
        _,
        diameter=diameter,
        channels=channels,
        model_type=model_type,
        pretrained_model=pretrained_model,
        cellpose_model_kwargs=cellpose_model_kwargs,
        **cellpose_eval_kwargs,
    )



import logging

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
from shapely.geometry import Polygon
from spatialdata import SpatialData
from spatialdata.models import ShapesModel
from spatialdata.transformations import get_transformation
from tqdm import tqdm

from ..shapes import ensure_polygon
from ..utils import to_intrinsic

log = logging.getLogger(__name__)


def solve_conflicts(
    cells: list[Polygon] | gpd.GeoDataFrame,
    threshold: float = 0.5,
    patch_indices: np.ndarray | None = None,
    return_indices: bool = False,
) -> gpd.GeoDataFrame | tuple[gpd.GeoDataFrame, np.ndarray]:
    """Resolve segmentation conflicts (i.e. overlap) after running segmentation on patches

    Args:
        cells: List of cell polygons
        threshold: When two cells are overlapping, we look at the area of intersection over the area of the smallest cell. If this value is higher than the `threshold`, the cells are merged
        patch_indices: Patch from which each cell belongs.
        return_indices: If `True`, returns also the cells indices. Merged cells have an index of -1.

    Returns:
        Array of resolved cells polygons. If `return_indices`, it also returns an array of cell indices.
    """
    cells = list(cells.geometry) if isinstance(cells, gpd.GeoDataFrame) else list(cells)
    n_cells = len(cells)
    resolved_indices = np.arange(n_cells)

    assert n_cells > 0, "No cells was segmented, cannot continue"

    tree = shapely.STRtree(cells)
    conflicts = tree.query(cells, predicate="intersects")

    if patch_indices is not None:
        conflicts = conflicts[:, patch_indices[conflicts[0]] != patch_indices[conflicts[1]]].T
    else:
        conflicts = conflicts[:, conflicts[0] != conflicts[1]].T

    for i1, i2 in tqdm(conflicts, desc="Resolving conflicts"):
        resolved_i1: int = resolved_indices[i1]
        resolved_i2: int = resolved_indices[i2]
        cell1, cell2 = cells[resolved_i1], cells[resolved_i2]

        intersection = cell1.intersection(cell2).area
        if intersection >= threshold * min(cell1.area, cell2.area):
            cell = ensure_polygon(cell1.union(cell2))
            assert not cell.is_empty, "Merged cell is empty"

            resolved_indices[np.isin(resolved_indices, [resolved_i1, resolved_i2])] = len(cells)
            cells.append(cell)

    unique_indices = np.unique(resolved_indices)
    unique_cells = gpd.GeoDataFrame(geometry=cells).iloc[unique_indices]

    if return_indices:
        return unique_cells, np.where(unique_indices < n_cells, unique_indices, -1)

    return unique_cells


def combine(
    sdata: SpatialData,
    elements: list[str | gpd.GeoDataFrame],
    key_added: str,
    threshold: float = 0.5,
):
    """Combine multiple segmentation boundaries into a single one.

    Example:
        On the example below, we run Cellpose twice, once for nuclei and once for tumor cells. We then combine the two segmentations into a single one.
        ```python
        import sopa

        sdata = sopa.io.toy_dataset(length=1000)
        sopa.make_image_patches(sdata)

        sopa.segmentation.cellpose(sdata, "DAPI", diameter=35, key_added="nuclei")
        sopa.segmentation.cellpose(sdata, ["DAPI", "CK"], diameter=35, key_added="tumor_cells")

        sopa.segmentation.combine(sdata, ["nuclei", "tumor_cells"], key_added="combined_cells")
        ```

    Args:
        sdata: A `SpatialData` object.
        elements: List of name of the keys in `sdata.shapes` to be combined (or directly a list of `GeoDataFrame`).
        key_added: The name of the new key to be added to `sdata.shapes`.
        threshold: When two cells are overlapping, we look at the area of intersection over the area of the smallest cell. If this value is higher than the `threshold`, the cells are merged
    """
    assert len(elements) > 1, "At least two elements must be provided to combine"

    elements: list[gpd.GeoDataFrame] = [
        element if isinstance(element, gpd.GeoDataFrame) else sdata.shapes[element] for element in elements
    ]

    reference = elements[0]
    intrinsic_elements = [reference] + [to_intrinsic(sdata, element, reference) for element in elements[1:]]

    combined_cells = list(pd.concat([element.geometry for element in intrinsic_elements], axis=0))
    combined_cells = solve_conflicts(combined_cells, threshold=threshold)

    combined_geo_df = ShapesModel.parse(combined_cells, transformations=get_transformation(reference, get_all=True))

    sdata.shapes[key_added] = combined_geo_df



import logging

import geopandas as gpd
import numpy as np
import pandas as pd
from anndata import AnnData
from scipy.spatial import Delaunay
from shapely.geometry import MultiPolygon, Polygon
from spatialdata import SpatialData

from ..constants import SopaKeys
from ._graph import Component
from .build import _check_has_delaunay

log = logging.getLogger(__name__)


def vectorize_niches(
    adata: AnnData | SpatialData,
    niche_key: str,
    buffer: int | str = "auto",
    perc_area_th: float = 0.05,
) -> gpd.GeoDataFrame:
    """Converts the niches to shapely polygons, and put into a `GeoDataFrame`. Note that each niche can appear multiple times, as they can be separated by other niches ; in this case, we call them different "components" of the same niche ID.

    Plot components:
        You can show niches components with GeoPandas
        ```py
        gdf = vectorize_niches(adata, niche_key)
        gdf.plot(column=niche_key)
        ```

    Args:
        adata: An `AnnData` object, or a `SpatialData object`
        niche_key: Key of `adata.obs` containing the niches
        buffer: Expansion radius applied on components. By default, `3 * mean_distance_neighbors`
        perc_area_th: For each niche, components whose area is less than `perc_area_th * max_component_area` will be removed

    Returns:
        A `GeoDataFrame` with geometries for each niche component. We also compute the area/perimeter/roundness of each component.
    """
    if isinstance(adata, SpatialData):
        adata = adata.tables[SopaKeys.TABLE]

    _check_has_delaunay(adata)
    data = {"geometry": [], niche_key: []}

    delaunay = Delaunay(adata.obsm["spatial"])
    connectivities = adata.obsp["spatial_connectivities"]
    values = adata.obs[niche_key].values

    keep = (
        (connectivities[delaunay.simplices[:, 0], delaunay.simplices[:, 1]].A1 == 1)
        & (connectivities[delaunay.simplices[:, 0], delaunay.simplices[:, 2]].A1 == 1)
        & (connectivities[delaunay.simplices[:, 1], delaunay.simplices[:, 2]].A1 == 1)
        & (values[delaunay.simplices[:, 0]] == values[delaunay.simplices[:, 1]])
        & (values[delaunay.simplices[:, 0]] == values[delaunay.simplices[:, 2]])
        & (values[delaunay.simplices[:, 0]] == values[delaunay.simplices[:, 2]])
    )  # Keep simplices that are in the original Delaunay graph, and which are not in between different value categories

    neighbors = np.where(np.isin(delaunay.neighbors, np.where(~keep)[0]), -1, delaunay.neighbors)

    simplices_to_visit = set(np.where(keep)[0])

    while simplices_to_visit:
        component = Component(adata, delaunay, neighbors)
        component.visit(simplices_to_visit)

        data["geometry"].append(component.polygon)
        data[niche_key].append(values[component.first_vertex_index()])

    gdf = gpd.GeoDataFrame(data)

    if buffer is not None and buffer != 0:
        gdf = _clean_components(adata, gdf, niche_key, buffer)

    gdf[SopaKeys.GEOMETRY_LENGTH] = gdf.length
    gdf[SopaKeys.GEOMETRY_AREA] = gdf.area
    gdf[SopaKeys.GEOMETRY_ROUNDNESS] = 4 * np.pi * gdf[SopaKeys.GEOMETRY_AREA] / gdf[SopaKeys.GEOMETRY_LENGTH] ** 2

    # Remove minor components (compared to the largest component of its corresponding niche)
    gdf = gdf[gdf.area >= gdf[niche_key].map(gdf.groupby(niche_key).area.max() * perc_area_th)]

    return gdf


def geometrize_niches(*args, **kwargs):
    """Alias for `vectorize_niches`"""
    log.warning(
        "The function `geometrize_niches` is deprecated and will be removed in a future version. "
        "Use `vectorize_niches` instead."
    )
    return vectorize_niches(*args, **kwargs)


def _clean_components(adata: AnnData, gdf: gpd.GeoDataFrame, niche_key: str, buffer: int | str) -> gpd.GeoDataFrame:
    data = {"geometry": [], niche_key: []}

    if buffer == "auto":
        buffer = 3 * adata.obsp["spatial_distances"].data.mean()

    for niche, group_gdf in gdf.groupby(niche_key):
        multi_polygon = MultiPolygon(group_gdf.geometry.values).buffer(buffer).buffer(-buffer)

        if isinstance(multi_polygon, Polygon):
            data["geometry"].append(multi_polygon)
            data[niche_key].append(niche)
            continue

        for geometry in multi_polygon.geoms:
            data["geometry"].append(geometry)
            data[niche_key].append(niche)

    return gpd.GeoDataFrame(data)


def niches_geometry_stats(
    adata: AnnData | SpatialData,
    niche_key: str,
    aggregation: str | list[str] = "min",
    key_added_suffix: str = "_distance_to_niche_",
    **vectorize_niches_kwargs: str,
) -> gpd.GeoDataFrame:
    """Computes statistics over niches geometries

    Details:
        - `n_components`: Number of connected component of a niche (a component is a group of neighbor cells with the same niche attribute)
        - `length`: Mean distance of the exterior/boundary of the components of a niche
        - `area`: Mean area of the components of a niche
        - `roundness`: Float value between 0 and 1. The higher the value, the closer to a circle. Computed via `4 * pi * area / length**2`
        - `mean_distance_to_niche_X`: mean distance to the niche (between the two closest points of the niches)

    Args:
        adata: An `AnnData` object, or a `SpatialData object`
        niche_key: Key of `adata.obs` containing the niches
        aggregation: Aggregation mode. Either one string such as `"min"`, or a list such as `["mean", "min"]`.
        key_added_suffix: Suffix added in the DataFrame columns. Defaults to "_distance_to_niche_".
        vectorize_niches_kwargs: Kwargs to the `sopa.spatial.vectorize_niches` function

    Returns:
        A `DataFrame` of shape `n_niches * n_statistics`
    """
    if isinstance(adata, SpatialData):
        adata = adata.tables[SopaKeys.TABLE]

    gdf = vectorize_niches(adata, niche_key, **vectorize_niches_kwargs)
    value_counts = gdf[niche_key].value_counts()

    assert len(gdf), "No niche geometry found, stats can't be computed"

    log.info(f"Computing pairwise distances between {len(gdf)} components")
    pairwise_distances: pd.DataFrame = gdf.geometry.apply(lambda g: gdf.distance(g))
    pairwise_distances[niche_key] = gdf[niche_key]

    if isinstance(aggregation, str):
        aggregation = [aggregation]

    for aggr in aggregation:
        df = pairwise_distances.groupby(niche_key).aggregate(aggr).T
        df.columns = [f"{aggr}{key_added_suffix}{c}" for c in df.columns]
        gdf[df.columns] = df

    df_stats: pd.DataFrame = gdf.groupby(niche_key)[gdf.columns[2:]].mean()
    df_stats.insert(0, SopaKeys.GEOMETRY_COUNT, value_counts)

    return df_stats



from pathlib import Path
from typing import Any, Literal

import xarray
from spatialdata import SpatialData
from spatialdata.models import Image2DModel
from spatialdata.transformations import Identity, Scale
from xarray import DataArray, Dataset, DataTree

from ...constants import SopaAttrs


def wsi(
    path: str | Path,
    chunks: tuple[int, int, int] = (3, 512, 512),
    as_image: bool = False,
    backend: Literal["tiffslide", "openslide", "slideio"] = "tiffslide",
) -> SpatialData | DataTree:
    """Read a WSI into a `SpatialData` object.

    !!! info "Supported backends"
        Multiple backends are supported to read WSIs. To use `openslide`, you will need to install `openslide-python` and `openslide-bin`.
        If you want to use `slideio`, you will need to install `slideio`. The `tiffslide` backend is supported out-of-the-box.

    Args:
        path: Path to the WSI
        chunks: Tuple representing the chunksize for the dimensions `(C, Y, X)`.
        as_image: If `True`, returns a, image instead of a `SpatialData` object
        backend: The library to use as a backend in order to load the WSI. One of: `"openslide"`, `"tiffslide"`, `"slideio"`.

    Returns:
        A `SpatialData` object with a multiscale 2D-image of shape `(C, Y, X)`, or just the DataTree if `as_image=True`
    """
    image_name, img, slide_metadata = _open_wsi(path, backend=backend)

    images = {}
    for level, key in enumerate(sorted(img.keys(), key=int)):
        suffix = key if key != "0" else ""

        scale_image = DataArray(
            img[key].transpose("S", f"Y{suffix}", f"X{suffix}"),
            dims=("c", "y", "x"),
        ).chunk(chunks)

        scale_factor = slide_metadata["level_downsamples"][level]

        scale_image = Image2DModel.parse(
            scale_image[:3, :, :],
            transformations={"global": _get_scale_transformation(scale_factor)},
            c_coords=("r", "g", "b"),
        )
        scale_image.coords["y"] = scale_factor * scale_image.coords["y"]
        scale_image.coords["x"] = scale_factor * scale_image.coords["x"]

        images[f"scale{key}"] = Dataset({"image": scale_image})

    multiscale_image = DataTree.from_dict(images)
    sdata = SpatialData(images={image_name: multiscale_image}, attrs={SopaAttrs.TISSUE_SEGMENTATION: image_name})
    sdata[image_name].attrs["metadata"] = slide_metadata
    sdata[image_name].attrs["backend"] = backend
    sdata[image_name].attrs["path"] = str(path)
    sdata[image_name].name = image_name

    if as_image:
        return multiscale_image

    return sdata


def _get_scale_transformation(scale_factor: float):
    if scale_factor == 1:
        return Identity()
    return Scale([scale_factor, scale_factor], axes=("x", "y"))


def wsi_autoscale(
    path: str | Path,
    image_model_kwargs: dict | None = None,
    backend: Literal["tiffslide", "openslide", "slideio"] = "tiffslide",
) -> SpatialData:
    """Read a WSI into a `SpatialData` object.

    Scales are generated automatically by `spatialdata` instead of using
    the default multiscales.

    Args:
        path: Path to the WSI
        image_model_kwargs: Kwargs provided to the `Image2DModel`
        backend: The library to use as a backend in order to load the WSI. One of: `"openslide"`, `"tiffslide"`, `"slideio"`.

    Returns:
        A `SpatialData` object with a 2D-image of shape `(C, Y, X)`
    """
    image_model_kwargs = _default_image_models_kwargs(image_model_kwargs)

    image_name, img, tiff_metadata = _open_wsi(path, backend=backend)

    img = img.rename_dims({"S": "c", "Y": "y", "X": "x"})

    multiscale_image = Image2DModel.parse(
        img["0"].transpose("c", "y", "x"),
        c_coords=("r", "g", "b"),
        **image_model_kwargs,
    )
    multiscale_image.attrs["metadata"] = tiff_metadata
    multiscale_image.attrs["backend"] = backend

    return SpatialData(images={image_name: multiscale_image}, attrs={SopaAttrs.TISSUE_SEGMENTATION: image_name})


def _default_image_models_kwargs(image_models_kwargs: dict | None) -> dict:
    image_models_kwargs = {} if image_models_kwargs is None else image_models_kwargs

    if "chunks" not in image_models_kwargs:
        image_models_kwargs["chunks"] = (3, 4096, 4096)

    if "scale_factors" not in image_models_kwargs:
        image_models_kwargs["scale_factors"] = [2, 2, 2, 2]

    return image_models_kwargs


def _open_wsi(
    path: str | Path, backend: Literal["tiffslide", "openslide", "slideio"] = "openslide"
) -> tuple[str, xarray.Dataset, Any, dict]:
    from ._wsi_reader import get_reader

    image_name = Path(path).stem

    reader = get_reader(backend)(path)

    zarr_store = reader.get_zarr_store()
    metadata = reader.get_metadata()

    zarr_img = xarray.open_zarr(zarr_store, consolidated=False, mask_and_scale=False)

    return image_name, zarr_img, metadata



import logging

import spatialdata
from spatialdata import SpatialData

from ..constants import VALID_DIMENSIONS
from ..utils import assert_is_integer_dtype, get_channel_names, get_spatial_image, is_valid_c_coords

log = logging.getLogger(__name__)


def sanity_check(sdata: SpatialData):
    assert len(sdata.images) > 0, "The spatialdata object has no image. Sopa is not designed for this."

    image = get_spatial_image(sdata)
    assert image.dims == VALID_DIMENSIONS, (
        f"Image must have the following three dimensions: {VALID_DIMENSIONS}. Found {image.dims}"
    )
    assert_is_integer_dtype(image.dtype)

    c_coords = get_channel_names(image)
    assert is_valid_c_coords(c_coords), f"Channel names must be strings, not {c_coords.dtype}"


def read_zarr_standardized(path: str) -> SpatialData:
    sdata = spatialdata.read_zarr(path)

    sanity_check(sdata)

    return sdata
